# encoding: utf-8

"""
Plot-related oxml objects.
"""

from __future__ import absolute_import, print_function, unicode_literals

from .datalabel import CT_DLbls
from ..simpletypes import (
    ST_BarDir,
    ST_BubbleScale,
    ST_GapAmount,
    ST_Grouping,
    ST_Overlap,
)
from ..xmlchemy import (
    BaseOxmlElement,
    OneAndOnlyOne,
    OptionalAttribute,
    ZeroOrOne,
    ZeroOrMore,
)


class BaseChartElement(BaseOxmlElement):
    """
    Base class for barChart, lineChart, and other plot elements.
    """

    @property
    def cat(self):
        """
        Return the `c:cat` element of the first series in this xChart, or
        |None| if not present.
        """
        cats = self.xpath("./c:ser[1]/c:cat")
        return cats[0] if cats else None

    @property
    def cat_pt_count(self):
        """
        Return the value of the `c:ptCount` descendent of this xChart
        element. Its parent can be one of three element types. This value
        represents the true number of (leaf) categories, although they might
        not all have a corresponding `c:pt` sibling; a category with no label
        does not get a `c:pt` element. Returns 0 if there is no `c:ptCount`
        descendent.
        """
        cat_ptCounts = self.xpath("./c:ser//c:cat//c:ptCount")
        if not cat_ptCounts:
            return 0
        return cat_ptCounts[0].val

    @property
    def cat_pts(self):
        """
        Return a sequence representing the `c:pt` elements under the `c:cat`
        element of the first series in this xChart element. A category having
        no value will have no corresponding `c:pt` element; |None| will
        appear in that position in such cases. Items appear in `idx` order.
        Only those in the first ``<c:lvl>`` element are included in the case
        of multi-level categories.
        """
        cat_pts = self.xpath("./c:ser[1]/c:cat//c:lvl[1]/c:pt")
        if not cat_pts:
            cat_pts = self.xpath("./c:ser[1]/c:cat//c:pt")

        cat_pt_dict = dict((pt.idx, pt) for pt in cat_pts)

        return [cat_pt_dict.get(idx, None) for idx in range(self.cat_pt_count)]

    @property
    def grouping_val(self):
        """
        Return the value of the ``./c:grouping{val=?}`` attribute, taking
        defaults into account when items are not present.
        """
        grouping = self.grouping
        if grouping is None:
            return ST_Grouping.STANDARD
        val = grouping.val
        if val is None:
            return ST_Grouping.STANDARD
        return val

    def iter_sers(self):
        """
        Generate each ``<c:ser>`` child element in this xChart in
        c:order/@val sequence (not document or c:idx order).
        """

        def ser_order(ser):
            return ser.order.val

        return (ser for ser in sorted(self.xpath("./c:ser"), key=ser_order))

    @property
    def sers(self):
        """
        Sequence of ``<c:ser>`` child elements in this xChart in c:order/@val
        sequence (not document or c:idx order).
        """
        return tuple(self.iter_sers())

    def _new_dLbls(self):
        return CT_DLbls.new_dLbls()


class CT_Area3DChart(BaseChartElement):
    """
    ``<c:area3DChart>`` element.
    """

    grouping = ZeroOrOne(
        "c:grouping",
        successors=(
            "c:varyColors",
            "c:ser",
            "c:dLbls",
            "c:dropLines",
            "c:gapDepth",
            "c:axId",
        ),
    )


class CT_AreaChart(BaseChartElement):
    """
    ``<c:areaChart>`` element.
    """

    _tag_seq = (
        "c:grouping",
        "c:varyColors",
        "c:ser",
        "c:dLbls",
        "c:dropLines",
        "c:axId",
        "c:extLst",
    )
    grouping = ZeroOrOne("c:grouping", successors=_tag_seq[1:])
    varyColors = ZeroOrOne("c:varyColors", successors=_tag_seq[2:])
    ser = ZeroOrMore("c:ser", successors=_tag_seq[3:])
    dLbls = ZeroOrOne("c:dLbls", successors=_tag_seq[4:])
    del _tag_seq


class CT_BarChart(BaseChartElement):
    """
    ``<c:barChart>`` element.
    """

    _tag_seq = (
        "c:barDir",
        "c:grouping",
        "c:varyColors",
        "c:ser",
        "c:dLbls",
        "c:gapWidth",
        "c:overlap",
        "c:serLines",
        "c:axId",
        "c:extLst",
    )
    barDir = OneAndOnlyOne("c:barDir")
    grouping = ZeroOrOne("c:grouping", successors=_tag_seq[2:])
    varyColors = ZeroOrOne("c:varyColors", successors=_tag_seq[3:])
    ser = ZeroOrMore("c:ser", successors=_tag_seq[4:])
    dLbls = ZeroOrOne("c:dLbls", successors=_tag_seq[5:])
    gapWidth = ZeroOrOne("c:gapWidth", successors=_tag_seq[6:])
    overlap = ZeroOrOne("c:overlap", successors=_tag_seq[7:])
    del _tag_seq

    @property
    def grouping_val(self):
        """
        Return the value of the ``./c:grouping{val=?}`` attribute, taking
        defaults into account when items are not present.
        """
        grouping = self.grouping
        if grouping is None:
            return ST_Grouping.CLUSTERED
        val = grouping.val
        if val is None:
            return ST_Grouping.CLUSTERED
        return val


class CT_BarDir(BaseOxmlElement):
    """
    ``<c:barDir>`` child of a barChart element, specifying the orientation of
    the bars, 'bar' if they are horizontal and 'col' if they are vertical.
    """

    val = OptionalAttribute("val", ST_BarDir, default=ST_BarDir.COL)


class CT_BubbleChart(BaseChartElement):
    """
    ``<c:bubbleChart>`` custom element class
    """

    _tag_seq = (
        "c:varyColors",
        "c:ser",
        "c:dLbls",
        "c:axId",
        "c:bubble3D",
        "c:bubbleScale",
        "c:showNegBubbles",
        "c:sizeRepresents",
        "c:axId",
        "c:extLst",
    )
    ser = ZeroOrMore("c:ser", successors=_tag_seq[2:])
    dLbls = ZeroOrOne("c:dLbls", successors=_tag_seq[3:])
    bubble3D = ZeroOrOne("c:bubble3D", successors=_tag_seq[5:])
    bubbleScale = ZeroOrOne("c:bubbleScale", successors=_tag_seq[6:])
    del _tag_seq


class CT_BubbleScale(BaseChartElement):
    """
    ``<c:bubbleScale>`` custom element class
    """

    val = OptionalAttribute("val", ST_BubbleScale, default=100)


class CT_DoughnutChart(BaseChartElement):
    """
    ``<c:doughnutChart>`` element.
    """

    _tag_seq = (
        "c:varyColors",
        "c:ser",
        "c:dLbls",
        "c:firstSliceAng",
        "c:holeSize",
        "c:extLst",
    )
    varyColors = ZeroOrOne("c:varyColors", successors=_tag_seq[1:])
    ser = ZeroOrMore("c:ser", successors=_tag_seq[2:])
    dLbls = ZeroOrOne("c:dLbls", successors=_tag_seq[3:])
    del _tag_seq


class CT_GapAmount(BaseOxmlElement):
    """
    ``<c:gapWidth>`` child of ``<c:barChart>`` element, also used for other
    purposes like error bars.
    """

    val = OptionalAttribute("val", ST_GapAmount, default=150)


class CT_Grouping(BaseOxmlElement):
    """
    ``<c:grouping>`` child of an xChart element, specifying a value like
    'clustered' or 'stacked'. Also used for variants with the same tag name
    like CT_BarGrouping.
    """

    val = OptionalAttribute("val", ST_Grouping)


class CT_LineChart(BaseChartElement):
    """
    ``<c:lineChart>`` custom element class
    """

    _tag_seq = (
        "c:grouping",
        "c:varyColors",
        "c:ser",
        "c:dLbls",
        "c:dropLines",
        "c:hiLowLines",
        "c:upDownBars",
        "c:marker",
        "c:smooth",
        "c:axId",
        "c:extLst",
    )
    grouping = ZeroOrOne("c:grouping", successors=(_tag_seq[1:]))
    varyColors = ZeroOrOne("c:varyColors", successors=_tag_seq[2:])
    ser = ZeroOrMore("c:ser", successors=_tag_seq[3:])
    dLbls = ZeroOrOne("c:dLbls", successors=(_tag_seq[4:]))
    del _tag_seq


class CT_Overlap(BaseOxmlElement):
    """
    ``<c:overlap>`` element specifying bar overlap as an integer percentage
    of bar width, in range -100 to 100.
    """

    val = OptionalAttribute("val", ST_Overlap, default=0)


class CT_PieChart(BaseChartElement):
    """
    ``<c:pieChart>`` custom element class
    """

    _tag_seq = ("c:varyColors", "c:ser", "c:dLbls", "c:firstSliceAng", "c:extLst")
    varyColors = ZeroOrOne("c:varyColors", successors=_tag_seq[1:])
    ser = ZeroOrMore("c:ser", successors=_tag_seq[2:])
    dLbls = ZeroOrOne("c:dLbls", successors=_tag_seq[3:])
    del _tag_seq


class CT_RadarChart(BaseChartElement):
    """
    ``<c:radarChart>`` custom element class
    """

    _tag_seq = (
        "c:radarStyle",
        "c:varyColors",
        "c:ser",
        "c:dLbls",
        "c:axId",
        "c:extLst",
    )
    varyColors = ZeroOrOne("c:varyColors", successors=_tag_seq[2:])
    ser = ZeroOrMore("c:ser", successors=_tag_seq[3:])
    dLbls = ZeroOrOne("c:dLbls", successors=(_tag_seq[4:]))
    del _tag_seq


class CT_ScatterChart(BaseChartElement):
    """
    ``<c:scatterChart>`` custom element class
    """

    _tag_seq = (
        "c:scatterStyle",
        "c:varyColors",
        "c:ser",
        "c:dLbls",
        "c:axId",
        "c:extLst",
    )
    varyColors = ZeroOrOne("c:varyColors", successors=_tag_seq[2:])
    ser = ZeroOrMore("c:ser", successors=_tag_seq[3:])
    del _tag_seq



# encoding: utf-8

"""
Chart builder and related objects.
"""

from __future__ import absolute_import, print_function, unicode_literals

from contextlib import contextmanager

from xlsxwriter import Workbook

from ..compat import BytesIO


class _BaseWorkbookWriter(object):
    """
    Base class for workbook writers, providing shared members.
    """

    def __init__(self, chart_data):
        super(_BaseWorkbookWriter, self).__init__()
        self._chart_data = chart_data

    @property
    def xlsx_blob(self):
        """
        Return the byte stream of an Excel file formatted as chart data for
        the category chart specified in the chart data object.
        """
        xlsx_file = BytesIO()
        with self._open_worksheet(xlsx_file) as (workbook, worksheet):
            self._populate_worksheet(workbook, worksheet)
        return xlsx_file.getvalue()

    @contextmanager
    def _open_worksheet(self, xlsx_file):
        """
        Enable XlsxWriter Worksheet object to be opened, operated on, and
        then automatically closed within a `with` statement. A filename or
        stream object (such as a ``BytesIO`` instance) is expected as
        *xlsx_file*.
        """
        workbook = Workbook(xlsx_file, {"in_memory": True})
        worksheet = workbook.add_worksheet()
        yield workbook, worksheet
        workbook.close()

    def _populate_worksheet(self, workbook, worksheet):
        """
        Must be overridden by each subclass to provide the particulars of
        writing the spreadsheet data.
        """
        raise NotImplementedError("must be provided by each subclass")


class CategoryWorkbookWriter(_BaseWorkbookWriter):
    """
    Determines Excel worksheet layout and can write an Excel workbook from
    a CategoryChartData object. Serves as the authority for Excel worksheet
    ranges.
    """

    @property
    def categories_ref(self):
        """
        The Excel worksheet reference to the categories for this chart (not
        including the column heading).
        """
        categories = self._chart_data.categories
        if categories.depth == 0:
            raise ValueError("chart data contains no categories")
        right_col = chr(ord("A") + categories.depth - 1)
        bottom_row = categories.leaf_count + 1
        return "Sheet1!$A$2:$%s$%d" % (right_col, bottom_row)

    def series_name_ref(self, series):
        """
        Return the Excel worksheet reference to the cell containing the name
        for *series*. This also serves as the column heading for the series
        values.
        """
        return "Sheet1!$%s$1" % self._series_col_letter(series)

    def values_ref(self, series):
        """
        The Excel worksheet reference to the values for this series (not
        including the column heading).
        """
        return "Sheet1!${col_letter}$2:${col_letter}${bottom_row}".format(
            **{
                "col_letter": self._series_col_letter(series),
                "bottom_row": len(series) + 1,
            }
        )

    @staticmethod
    def _column_reference(column_number):
        """Return str Excel column reference like 'BQ' for *column_number*.

        *column_number* is an int in the range 1-16384 inclusive, where
        1 maps to column 'A'.
        """
        if column_number < 1 or column_number > 16384:
            raise ValueError("column_number must be in range 1-16384")

        # ---Work right-to-left, one order of magnitude at a time. Note there
        #    is no zero representation in Excel address scheme, so this is
        #    not just a conversion to base-26---

        col_ref = ""
        while column_number:
            remainder = column_number % 26
            if remainder == 0:
                remainder = 26

            col_letter = chr(ord("A") + remainder - 1)
            col_ref = col_letter + col_ref

            # ---Advance to next order of magnitude or terminate loop. The
            # minus-one in this expression reflects the fact the next lower
            # order of magnitude has a minumum value of 1 (not zero). This is
            # essentially the complement to the "if it's 0 make it 26' step
            # above.---
            column_number = (column_number - 1) // 26

        return col_ref

    def _populate_worksheet(self, workbook, worksheet):
        """
        Write the chart data contents to *worksheet* in category chart
        layout. Write categories starting in the first column starting in
        the second row, and proceeding one column per category level (for
        charts having multi-level categories). Write series as columns
        starting in the next following column, placing the series title in
        the first cell.
        """
        self._write_categories(workbook, worksheet)
        self._write_series(workbook, worksheet)

    def _series_col_letter(self, series):
        """
        The letter of the Excel worksheet column in which the data for a
        series appears.
        """
        column_number = 1 + series.categories.depth + series.index
        return self._column_reference(column_number)

    def _write_categories(self, workbook, worksheet):
        """
        Write the categories column(s) to *worksheet*. Categories start in
        the first column starting in the second row, and proceeding one
        column per category level (for charts having multi-level categories).
        A date category is formatted as a date. All others are formatted
        `General`.
        """
        categories = self._chart_data.categories
        num_format = workbook.add_format({"num_format": categories.number_format})
        depth = categories.depth
        for idx, level in enumerate(categories.levels):
            col = depth - idx - 1
            self._write_cat_column(worksheet, col, level, num_format)

    def _write_cat_column(self, worksheet, col, level, num_format):
        """
        Write a category column defined by *level* to *worksheet* at offset
        *col* and formatted with *num_format*.
        """
        worksheet.set_column(col, col, 10)  # wide enough for a date
        for off, name in level:
            row = off + 1
            worksheet.write(row, col, name, num_format)

    def _write_series(self, workbook, worksheet):
        """
        Write the series column(s) to *worksheet*. Series start in the column
        following the last categories column, placing the series title in the
        first cell.
        """
        col_offset = self._chart_data.categories.depth
        for idx, series in enumerate(self._chart_data):
            num_format = workbook.add_format({"num_format": series.number_format})
            series_col = idx + col_offset
            worksheet.write(0, series_col, series.name)
            worksheet.write_column(1, series_col, series.values, num_format)


class XyWorkbookWriter(_BaseWorkbookWriter):
    """
    Determines Excel worksheet layout and can write an Excel workbook from XY
    chart data. Serves as the authority for Excel worksheet ranges.
    """

    def series_name_ref(self, series):
        """
        Return the Excel worksheet reference to the cell containing the name
        for *series*. This also serves as the column heading for the series
        Y values.
        """
        row = self.series_table_row_offset(series) + 1
        return "Sheet1!$B$%d" % row

    def series_table_row_offset(self, series):
        """
        Return the number of rows preceding the data table for *series* in
        the Excel worksheet.
        """
        title_and_spacer_rows = series.index * 2
        data_point_rows = series.data_point_offset
        return title_and_spacer_rows + data_point_rows

    def x_values_ref(self, series):
        """
        The Excel worksheet reference to the X values for this chart (not
        including the column label).
        """
        top_row = self.series_table_row_offset(series) + 2
        bottom_row = top_row + len(series) - 1
        return "Sheet1!$A$%d:$A$%d" % (top_row, bottom_row)

    def y_values_ref(self, series):
        """
        The Excel worksheet reference to the Y values for this chart (not
        including the column label).
        """
        top_row = self.series_table_row_offset(series) + 2
        bottom_row = top_row + len(series) - 1
        return "Sheet1!$B$%d:$B$%d" % (top_row, bottom_row)

    def _populate_worksheet(self, workbook, worksheet):
        """
        Write chart data contents to *worksheet* in the standard XY chart
        layout. Write the data for each series to a separate two-column
        table, X values in column A and Y values in column B. Place the
        series label in the first (heading) cell of the column.
        """
        chart_num_format = workbook.add_format(
            {"num_format": self._chart_data.number_format}
        )
        for series in self._chart_data:
            series_num_format = workbook.add_format(
                {"num_format": series.number_format}
            )
            offset = self.series_table_row_offset(series)
            # write X values
            worksheet.write_column(offset + 1, 0, series.x_values, chart_num_format)
            # write Y values
            worksheet.write(offset, 1, series.name)
            worksheet.write_column(offset + 1, 1, series.y_values, series_num_format)


class BubbleWorkbookWriter(XyWorkbookWriter):
    """
    Service object that knows how to write an Excel workbook from bubble
    chart data.
    """

    def bubble_sizes_ref(self, series):
        """
        The Excel worksheet reference to the range containing the bubble
        sizes for *series* (not including the column heading cell).
        """
        top_row = self.series_table_row_offset(series) + 2
        bottom_row = top_row + len(series) - 1
        return "Sheet1!$C$%d:$C$%d" % (top_row, bottom_row)

    def _populate_worksheet(self, workbook, worksheet):
        """
        Write chart data contents to *worksheet* in the bubble chart layout.
        Write the data for each series to a separate three-column table with
        X values in column A, Y values in column B, and bubble sizes in
        column C. Place the series label in the first (heading) cell of the
        values column.
        """
        chart_num_format = workbook.add_format(
            {"num_format": self._chart_data.number_format}
        )
        for series in self._chart_data:
            series_num_format = workbook.add_format(
                {"num_format": series.number_format}
            )
            offset = self.series_table_row_offset(series)
            # write X values
            worksheet.write_column(offset + 1, 0, series.x_values, chart_num_format)
            # write Y values
            worksheet.write(offset, 1, series.name)
            worksheet.write_column(offset + 1, 1, series.y_values, series_num_format)
            # write bubble sizes
            worksheet.write(offset, 2, "Size")
            worksheet.write_column(offset + 1, 2, series.bubble_sizes, chart_num_format)



# encoding: utf-8

"""Unit-test suite for `pptx.parts.slide` module."""

import pytest

from pptx.chart.data import ChartData
from pptx.enum.base import EnumValue
from pptx.enum.shapes import PROG_ID
from pptx.media import Video
from pptx.opc.constants import CONTENT_TYPE as CT, RELATIONSHIP_TYPE as RT
from pptx.opc.package import Part
from pptx.opc.packuri import PackURI
from pptx.oxml.slide import CT_NotesMaster, CT_NotesSlide, CT_Slide
from pptx.oxml.theme import CT_OfficeStyleSheet
from pptx.package import Package
from pptx.parts.chart import ChartPart
from pptx.parts.embeddedpackage import EmbeddedPackagePart
from pptx.parts.image import Image, ImagePart
from pptx.parts.media import MediaPart
from pptx.parts.presentation import PresentationPart
from pptx.parts.slide import (
    BaseSlidePart,
    NotesMasterPart,
    NotesSlidePart,
    SlideLayoutPart,
    SlideMasterPart,
    SlidePart,
)
from pptx.slide import NotesMaster, NotesSlide, Slide, SlideLayout, SlideMaster

from ..unitutil.cxml import element
from ..unitutil.file import absjoin, test_file_dir
from ..unitutil.mock import (
    call,
    class_mock,
    initializer_mock,
    instance_mock,
    method_mock,
    property_mock,
)


class DescribeBaseSlidePart(object):
    def it_knows_its_name(self, name_fixture):
        base_slide, expected_value = name_fixture
        assert base_slide.name == expected_value

    def it_can_get_a_related_image_by_rId(self, get_image_fixture):
        slide, rId, image_ = get_image_fixture
        assert slide.get_image(rId) is image_

    def it_can_add_an_image_part(self, image_part_fixture):
        slide, image_file, image_part_, rId_ = image_part_fixture

        image_part, rId = slide.get_or_add_image_part(image_file)

        slide._package.get_or_add_image_part.assert_called_once_with(image_file)
        slide.relate_to.assert_called_once_with(image_part_, RT.IMAGE)
        assert image_part is image_part_
        assert rId is rId_

    # fixtures -------------------------------------------------------

    @pytest.fixture
    def get_image_fixture(self, related_parts_prop_, image_part_, image_):
        slide = BaseSlidePart(None, None, None, None)
        rId = "rId42"
        related_parts_prop_.return_value = {rId: image_part_}
        image_part_.image = image_
        return slide, rId, image_

    @pytest.fixture
    def image_part_fixture(self, partname_, package_, image_part_, relate_to_):
        slide = BaseSlidePart(partname_, None, None, package_)
        image_file, rId = "foobar.png", "rId6"
        package_.get_or_add_image_part.return_value = image_part_
        relate_to_.return_value = rId
        return slide, image_file, image_part_, rId

    @pytest.fixture
    def name_fixture(self):
        sld_cxml, expected_value = "p:sld/p:cSld{name=Foobar}", "Foobar"
        sld = element(sld_cxml)
        base_slide = BaseSlidePart(None, None, sld, None)
        return base_slide, expected_value

    # fixture components ---------------------------------------------

    @pytest.fixture
    def image_(self, request):
        return instance_mock(request, Image)

    @pytest.fixture
    def image_part_(self, request):
        return instance_mock(request, ImagePart)

    @pytest.fixture
    def package_(self, request):
        return instance_mock(request, Package)

    @pytest.fixture
    def partname_(self):
        return PackURI("/foo/bar.xml")

    @pytest.fixture
    def relate_to_(self, request):
        return method_mock(request, BaseSlidePart, "relate_to")

    @pytest.fixture
    def related_parts_prop_(self, request):
        return property_mock(request, BaseSlidePart, "related_parts")


class DescribeNotesMasterPart(object):
    def it_can_create_a_notes_master_part(self, create_fixture):
        package_, theme_part_, notes_master_part_ = create_fixture

        notes_master_part = NotesMasterPart.create_default(package_)

        NotesMasterPart._new.assert_called_once_with(package_)
        NotesMasterPart._new_theme_part.assert_called_once_with(package_)
        notes_master_part.relate_to.assert_called_once_with(theme_part_, RT.THEME)
        assert notes_master_part is notes_master_part_

    def it_provides_access_to_its_notes_master(self, notes_master_fixture):
        notes_master_part, NotesMaster_ = notes_master_fixture[:2]
        notesMaster, notes_master_ = notes_master_fixture[2:]

        notes_master = notes_master_part.notes_master

        NotesMaster_.assert_called_once_with(notesMaster, notes_master_part)
        assert notes_master is notes_master_

    def it_creates_a_new_notes_master_part_to_help(self, new_fixture):
        package_, NotesMasterPart_, partname = new_fixture[:3]
        notesMaster_, notes_master_part_ = new_fixture[3:]

        notes_master_part = NotesMasterPart._new(package_)

        CT_NotesMaster.new_default.assert_called_once_with()
        NotesMasterPart_.assert_called_once_with(
            partname, CT.PML_NOTES_MASTER, notesMaster_, package_
        )
        assert notes_master_part is notes_master_part_

    def it_creates_a_new_theme_part_to_help(self, theme_fixture):
        package_, pn_tmpl, XmlPart_, partname = theme_fixture[:4]
        theme_elm_, theme_part_ = theme_fixture[4:]

        theme_part = NotesMasterPart._new_theme_part(package_)

        package_.next_partname.assert_called_once_with(pn_tmpl)
        CT_OfficeStyleSheet.new_default.assert_called_once_with()
        XmlPart_.assert_called_once_with(partname, CT.OFC_THEME, theme_elm_, package_)
        assert theme_part is theme_part_

    # fixtures -------------------------------------------------------

    @pytest.fixture
    def create_fixture(
        self, package_, theme_part_, notes_master_part_, _new_, _new_theme_part_
    ):
        return package_, theme_part_, notes_master_part_

    @pytest.fixture
    def new_fixture(
        self, package_, NotesMasterPart_, notesMaster_, notes_master_part_, new_default_
    ):
        partname = PackURI("/ppt/notesMasters/notesMaster1.xml")
        return (package_, NotesMasterPart_, partname, notesMaster_, notes_master_part_)

    @pytest.fixture
    def notes_master_fixture(self, NotesMaster_, notes_master_):
        notesMaster = element("p:notesMaster")
        notes_master_part = NotesMasterPart(None, None, notesMaster, None)
        return notes_master_part, NotesMaster_, notesMaster, notes_master_

    @pytest.fixture
    def theme_fixture(
        self, package_, XmlPart_, theme_elm_, theme_part_, theme_new_default_
    ):
        pn_tmpl = "/ppt/theme/theme%d.xml"
        partname = PackURI("/ppt/theme/theme2.xml")
        package_.next_partname.return_value = partname
        return (package_, pn_tmpl, XmlPart_, partname, theme_elm_, theme_part_)

    # fixture components ---------------------------------------------

    @pytest.fixture
    def _new_(self, request, notes_master_part_):
        return method_mock(
            request, NotesMasterPart, "_new", return_value=notes_master_part_
        )

    @pytest.fixture
    def new_default_(self, request, notesMaster_):
        return method_mock(
            request, CT_NotesMaster, "new_default", return_value=notesMaster_
        )

    @pytest.fixture
    def _new_theme_part_(self, request, theme_part_):
        return method_mock(
            request, NotesMasterPart, "_new_theme_part", return_value=theme_part_
        )

    @pytest.fixture
    def NotesMaster_(self, request, notes_master_):
        return class_mock(
            request, "pptx.parts.slide.NotesMaster", return_value=notes_master_
        )

    @pytest.fixture
    def NotesMasterPart_(self, request, notes_master_part_):
        return class_mock(
            request, "pptx.parts.slide.NotesMasterPart", return_value=notes_master_part_
        )

    @pytest.fixture
    def notesMaster_(self, request):
        return instance_mock(request, CT_NotesMaster)

    @pytest.fixture
    def notes_master_(self, request):
        return instance_mock(request, NotesMaster)

    @pytest.fixture
    def notes_master_part_(self, request):
        return instance_mock(request, NotesMasterPart)

    @pytest.fixture
    def package_(self, request):
        return instance_mock(request, Package)

    @pytest.fixture
    def theme_elm_(self, request):
        return instance_mock(request, CT_OfficeStyleSheet)

    @pytest.fixture
    def theme_new_default_(self, request, theme_elm_):
        return method_mock(
            request, CT_OfficeStyleSheet, "new_default", return_value=theme_elm_
        )

    @pytest.fixture
    def theme_part_(self, request):
        return instance_mock(request, Part)

    @pytest.fixture
    def XmlPart_(self, request, theme_part_):
        return class_mock(request, "pptx.parts.slide.XmlPart", return_value=theme_part_)


class DescribeNotesSlidePart(object):
    def it_can_create_a_notes_slide_part(self, new_fixture):
        package_, slide_part_, notes_master_part_ = new_fixture[:3]
        notes_slide_, notes_master_, notes_slide_part_ = new_fixture[3:]

        notes_slide_part = NotesSlidePart.new(package_, slide_part_)

        NotesSlidePart._add_notes_slide_part.assert_called_once_with(
            package_, slide_part_, notes_master_part_
        )
        notes_slide_.clone_master_placeholders.assert_called_once_with(notes_master_)
        assert notes_slide_part is notes_slide_part_

    def it_provides_access_to_the_notes_master(self, notes_master_fixture):
        notes_slide_part, notes_master_ = notes_master_fixture
        notes_master = notes_slide_part.notes_master
        notes_slide_part.part_related_by.assert_called_once_with(
            notes_slide_part, RT.NOTES_MASTER
        )
        assert notes_master is notes_master_

    def it_provides_access_to_its_notes_slide(self, notes_slide_fixture):
        notes_slide_part, NotesSlide_, notes, notes_slide_ = notes_slide_fixture
        notes_slide = notes_slide_part.notes_slide
        NotesSlide_.assert_called_once_with(notes, notes_slide_part)
        assert notes_slide is notes_slide_

    def it_adds_a_notes_slide_part_to_help(self, add_fixture):
        package_, slide_part_, notes_master_part_ = add_fixture[:3]
        notes_slide_part_, NotesSlidePart_, partname = add_fixture[3:6]
        content_type, notes, calls = add_fixture[6:]

        notes_slide_part = NotesSlidePart._add_notes_slide_part(
            package_, slide_part_, notes_master_part_
        )

        package_.next_partname.assert_called_once_with(
            "/ppt/notesSlides/notesSlide%d.xml"
        )
        CT_NotesSlide.new.assert_called_once_with()
        NotesSlidePart_.assert_called_once_with(partname, content_type, notes, package_)
        assert notes_slide_part_.relate_to.call_args_list == calls
        assert notes_slide_part is notes_slide_part_

    # fixtures -------------------------------------------------------

    @pytest.fixture
    def add_fixture(
        self,
        package_,
        slide_part_,
        notes_master_part_,
        notes_slide_part_,
        NotesSlidePart_,
        new_,
    ):
        partname = PackURI("/ppt/notesSlides/notesSlide42.xml")
        content_type = CT.PML_NOTES_SLIDE
        notes = element("p:notes")
        calls = [call(notes_master_part_, RT.NOTES_MASTER), call(slide_part_, RT.SLIDE)]
        package_.next_partname.return_value = partname
        new_.return_value = notes
        return (
            package_,
            slide_part_,
            notes_master_part_,
            notes_slide_part_,
            NotesSlidePart_,
            partname,
            content_type,
            notes,
            calls,
        )

    @pytest.fixture
    def new_fixture(
        self,
        package_,
        slide_part_,
        notes_master_part_,
        notes_slide_,
        notes_master_,
        notes_slide_part_,
        _add_notes_slide_part_,
        presentation_part_,
    ):
        package_.presentation_part = presentation_part_
        presentation_part_.notes_master_part = notes_master_part_
        notes_slide_part_.notes_slide = notes_slide_
        notes_master_part_.notes_master = notes_master_
        return (
            package_,
            slide_part_,
            notes_master_part_,
            notes_slide_,
            notes_master_,
            notes_slide_part_,
        )

    @pytest.fixture
    def notes_master_fixture(self, notes_master_, part_related_by_, notes_master_part_):
        notes_slide_part = NotesSlidePart(None, None, None, None)
        part_related_by_.return_value = notes_master_part_
        notes_master_part_.notes_master = notes_master_
        return notes_slide_part, notes_master_

    @pytest.fixture
    def notes_slide_fixture(self, NotesSlide_, notes_slide_):
        notes = element("p:notes")
        notes_slide_part = NotesSlidePart(None, None, notes, None)
        return notes_slide_part, NotesSlide_, notes, notes_slide_

    # fixture components ---------------------------------------------

    @pytest.fixture
    def _add_notes_slide_part_(self, request, notes_slide_part_):
        return method_mock(
            request,
            NotesSlidePart,
            "_add_notes_slide_part",
            return_value=notes_slide_part_,
        )

    @pytest.fixture
    def new_(self, request):
        return method_mock(request, CT_NotesSlide, "new")

    @pytest.fixture
    def notes_master_(self, request):
        return instance_mock(request, NotesMaster)

    @pytest.fixture
    def notes_master_part_(self, request):
        return instance_mock(request, NotesMasterPart)

    @pytest.fixture
    def NotesSlide_(self, request, notes_slide_):
        return class_mock(
            request, "pptx.parts.slide.NotesSlide", return_value=notes_slide_
        )

    @pytest.fixture
    def notes_slide_(self, request):
        return instance_mock(request, NotesSlide)

    @pytest.fixture
    def NotesSlidePart_(self, request, notes_slide_part_):
        return class_mock(
            request, "pptx.parts.slide.NotesSlidePart", return_value=notes_slide_part_
        )

    @pytest.fixture
    def notes_slide_part_(self, request):
        return instance_mock(request, NotesSlidePart)

    @pytest.fixture
    def part_related_by_(self, request):
        return method_mock(request, NotesSlidePart, "part_related_by", autospec=True)

    @pytest.fixture
    def presentation_part_(self, request):
        return instance_mock(request, PresentationPart)

    @pytest.fixture
    def package_(self, request):
        return instance_mock(request, Package)

    @pytest.fixture
    def slide_part_(self, request):
        return instance_mock(request, SlidePart)


class DescribeSlidePart(object):
    """Unit-test suite for `pptx.parts.slide.SlidePart` objects."""

    def it_knows_its_slide_id(self, slide_id_fixture):
        slide_part, presentation_part_, slide_id = slide_id_fixture
        _slide_id = slide_part.slide_id
        presentation_part_.slide_id.assert_called_once_with(slide_part)
        assert _slide_id is slide_id

    def it_knows_whether_it_has_a_notes_slide(self, has_notes_slide_fixture):
        slide_part, expected_value = has_notes_slide_fixture
        value = slide_part.has_notes_slide
        slide_part.part_related_by.assert_called_once_with(slide_part, RT.NOTES_SLIDE)
        assert value is expected_value

    def it_can_add_a_chart_part(self, request, package_, relate_to_):
        chart_part_ = instance_mock(request, ChartPart)
        ChartPart_ = class_mock(request, "pptx.parts.slide.ChartPart")
        ChartPart_.new.return_value = chart_part_
        chart_type_ = instance_mock(request, EnumValue)
        chart_data_ = instance_mock(request, ChartData)
        relate_to_.return_value = "rId42"
        slide_part = SlidePart(None, None, None, package_)

        _rId = slide_part.add_chart_part(chart_type_, chart_data_)

        ChartPart_.new.assert_called_once_with(chart_type_, chart_data_, package_)
        relate_to_.assert_called_once_with(slide_part, chart_part_, RT.CHART)
        assert _rId == "rId42"

    @pytest.mark.parametrize(
        "prog_id, rel_type",
        (
            (PROG_ID.DOCX, RT.PACKAGE),
            (PROG_ID.PPTX, RT.PACKAGE),
            (PROG_ID.XLSX, RT.PACKAGE),
            ("Foo.Bar.18", RT.OLE_OBJECT),
        ),
    )
    def it_can_add_an_embedded_ole_object_part(
        self, request, package_, relate_to_, prog_id, rel_type
    ):
        _blob_from_file_ = method_mock(
            request, SlidePart, "_blob_from_file", autospec=True, return_value=b"012345"
        )
        embedded_package_part_ = instance_mock(request, EmbeddedPackagePart)
        EmbeddedPackagePart_ = class_mock(
            request, "pptx.parts.slide.EmbeddedPackagePart"
        )
        EmbeddedPackagePart_.factory.return_value = embedded_package_part_
        relate_to_.return_value = "rId9"
        slide_part = SlidePart(None, None, None, package_)

        _rId = slide_part.add_embedded_ole_object_part(prog_id, "workbook.xlsx")

        _blob_from_file_.assert_called_once_with(slide_part, "workbook.xlsx")
        EmbeddedPackagePart_.factory.assert_called_once_with(
            prog_id, b"012345", package_
        )
        relate_to_.assert_called_once_with(slide_part, embedded_package_part_, rel_type)
        assert _rId == "rId9"

    def it_can_get_or_add_a_video_part(self, package_, video_, relate_to_, media_part_):
        media_rId, video_rId = "rId1", "rId2"
        package_.get_or_add_media_part.return_value = media_part_
        relate_to_.side_effect = [media_rId, video_rId]
        slide_part = SlidePart(None, None, None, package_)

        result = slide_part.get_or_add_video_media_part(video_)

        package_.get_or_add_media_part.assert_called_once_with(video_)
        assert relate_to_.call_args_list == [
            call(slide_part, media_part_, RT.MEDIA),
            call(slide_part, media_part_, RT.VIDEO),
        ]
        assert result == (media_rId, video_rId)

    def it_can_create_a_new_slide_part(self, request, package_, relate_to_):
        partname = PackURI("/foobar.xml")
        SlidePart_init_ = initializer_mock(request, SlidePart)
        slide_layout_part_ = instance_mock(request, SlideLayoutPart)
        CT_Slide_ = class_mock(request, "pptx.parts.slide.CT_Slide")
        CT_Slide_.new.return_value = sld = element("c:sld")

        slide_part = SlidePart.new(partname, package_, slide_layout_part_)

        SlidePart_init_.assert_called_once_with(partname, CT.PML_SLIDE, sld, package_)
        slide_part.relate_to.assert_called_once_with(
            slide_part, slide_layout_part_, RT.SLIDE_LAYOUT
        )
        assert isinstance(slide_part, SlidePart)

    def it_provides_access_to_its_slide(self, slide_fixture):
        slide_part, Slide_, sld, slide_ = slide_fixture
        slide = slide_part.slide
        Slide_.assert_called_once_with(sld, slide_part)
        assert slide is slide_

    def it_provides_access_to_the_slide_layout(self, layout_fixture):
        slide_part, slide_layout_ = layout_fixture
        slide_layout = slide_part.slide_layout
        slide_part.part_related_by.assert_called_once_with(slide_part, RT.SLIDE_LAYOUT)
        assert slide_layout is slide_layout_

    def it_knows_the_minimal_element_xml_for_a_slide(self):
        path = absjoin(test_file_dir, "minimal_slide.xml")
        sld = CT_Slide.new()
        with open(path, "r") as f:
            expected_xml = f.read()
        assert sld.xml == expected_xml

    def it_gets_its_notes_slide_to_help(self, notes_slide_fixture):
        slide_part, calls, notes_slide_ = notes_slide_fixture

        notes_slide = slide_part.notes_slide

        slide_part.part_related_by.assert_called_once_with(slide_part, RT.NOTES_SLIDE)
        assert slide_part._add_notes_slide_part.call_args_list == calls
        assert notes_slide is notes_slide_

    def it_adds_a_notes_slide_part_to_help(
        self, package_, NotesSlidePart_, notes_slide_part_, relate_to_
    ):
        NotesSlidePart_.new.return_value = notes_slide_part_
        slide_part = SlidePart(None, None, None, package_)

        notes_slide_part = slide_part._add_notes_slide_part()

        assert notes_slide_part is notes_slide_part_
        NotesSlidePart_.new.assert_called_once_with(package_, slide_part)
        relate_to_.assert_called_once_with(slide_part, notes_slide_part, RT.NOTES_SLIDE)

    # fixtures -------------------------------------------------------

    @pytest.fixture(params=[True, False])
    def has_notes_slide_fixture(self, request, part_related_by_):
        has_notes_slide = request.param
        slide_part = SlidePart(None, None, None, None)
        part_related_by_.side_effect = None if has_notes_slide else KeyError
        expected_value = has_notes_slide
        return slide_part, expected_value

    @pytest.fixture
    def layout_fixture(self, slide_layout_, part_related_by_):
        slide_part = SlidePart(None, None, None, None)
        part_related_by_.return_value.slide_layout = slide_layout_
        return slide_part, slide_layout_

    @pytest.fixture(params=[True, False])
    def notes_slide_fixture(
        self,
        request,
        notes_slide_,
        part_related_by_,
        _add_notes_slide_part_,
        notes_slide_part_,
    ):
        has_notes_slide = request.param
        slide_part = SlidePart(None, None, None, None)
        part_related_by_.return_value = notes_slide_part_
        add_calls = []
        if not has_notes_slide:
            part_related_by_.side_effect = KeyError
            add_calls.append(call(slide_part))
        notes_slide_part_.notes_slide = notes_slide_
        return slide_part, add_calls, notes_slide_

    @pytest.fixture
    def slide_fixture(self, Slide_, slide_):
        sld = element("p:sld")
        slide_part = SlidePart(None, None, sld, None)
        return slide_part, Slide_, sld, slide_

    @pytest.fixture
    def slide_id_fixture(self, package_, presentation_part_):
        slide_part = SlidePart(None, None, None, package_)
        slide_id = 256
        package_.presentation_part = presentation_part_
        presentation_part_.slide_id.return_value = slide_id
        return slide_part, presentation_part_, slide_id

    # fixture components ---------------------------------------------

    @pytest.fixture
    def _add_notes_slide_part_(self, request, notes_slide_part_):
        return method_mock(
            request,
            SlidePart,
            "_add_notes_slide_part",
            return_value=notes_slide_part_,
            autospec=True,
        )

    @pytest.fixture
    def media_part_(self, request):
        return instance_mock(request, MediaPart)

    @pytest.fixture
    def notes_slide_(self, request):
        return instance_mock(request, NotesSlide)

    @pytest.fixture
    def NotesSlidePart_(self, request, notes_slide_part_):
        return class_mock(
            request, "pptx.parts.slide.NotesSlidePart", return_value=notes_slide_part_
        )

    @pytest.fixture
    def notes_slide_part_(self, request):
        return instance_mock(request, NotesSlidePart)

    @pytest.fixture
    def package_(self, request):
        return instance_mock(request, Package)

    @pytest.fixture
    def part_related_by_(self, request):
        return method_mock(request, SlidePart, "part_related_by", autospec=True)

    @pytest.fixture
    def presentation_part_(self, request):
        return instance_mock(request, PresentationPart)

    @pytest.fixture
    def relate_to_(self, request):
        return method_mock(request, SlidePart, "relate_to", autospec=True)

    @pytest.fixture
    def Slide_(self, request, slide_):
        return class_mock(request, "pptx.parts.slide.Slide", return_value=slide_)

    @pytest.fixture
    def slide_(self, request):
        return instance_mock(request, Slide)

    @pytest.fixture
    def slide_layout_(self, request):
        return instance_mock(request, SlideLayout)

    @pytest.fixture
    def video_(self, request):
        return instance_mock(request, Video)


class DescribeSlideLayoutPart(object):
    def it_provides_access_to_its_slide_master(self, master_fixture):
        slide_layout_part, part_related_by_, slide_master_ = master_fixture
        slide_master = slide_layout_part.slide_master
        part_related_by_.assert_called_once_with(slide_layout_part, RT.SLIDE_MASTER)
        assert slide_master is slide_master_

    def it_provides_access_to_its_slide_layout(self, layout_fixture):
        slide_layout_part, SlideLayout_ = layout_fixture[:2]
        sldLayout, slide_layout_ = layout_fixture[2:]
        slide_layout = slide_layout_part.slide_layout
        SlideLayout_.assert_called_once_with(sldLayout, slide_layout_part)
        assert slide_layout is slide_layout_

    # fixtures -------------------------------------------------------

    @pytest.fixture
    def layout_fixture(self, SlideLayout_, slide_layout_):
        sldLayout = element("p:sldLayout")
        slide_layout_part = SlideLayoutPart(None, None, sldLayout)
        return slide_layout_part, SlideLayout_, sldLayout, slide_layout_

    @pytest.fixture
    def master_fixture(self, part_related_by_, slide_master_, slide_master_part_):
        slide_layout_part = SlideLayoutPart(None, None, None, None)
        part_related_by_.return_value = slide_master_part_
        slide_master_part_.slide_master = slide_master_
        return slide_layout_part, part_related_by_, slide_master_

    # fixture components -----------------------------------

    @pytest.fixture
    def part_related_by_(self, request):
        return method_mock(request, SlideLayoutPart, "part_related_by", autospec=True)

    @pytest.fixture
    def SlideLayout_(self, request, slide_layout_):
        return class_mock(
            request, "pptx.parts.slide.SlideLayout", return_value=slide_layout_
        )

    @pytest.fixture
    def slide_layout_(self, request):
        return instance_mock(request, SlideLayout)

    @pytest.fixture
    def slide_master_(self, request):
        return instance_mock(request, SlideMaster)

    @pytest.fixture
    def slide_master_part_(self, request):
        return instance_mock(request, SlideMasterPart)


class DescribeSlideMasterPart(object):
    def it_provides_access_to_its_slide_master(self, master_fixture):
        slide_master_part, SlideMaster_, sldMaster, slide_master_ = master_fixture
        slide_master = slide_master_part.slide_master
        SlideMaster_.assert_called_once_with(sldMaster, slide_master_part)
        assert slide_master is slide_master_

    def it_provides_access_to_a_related_slide_layout(self, related_fixture):
        slide_master_part, rId, getitem_, slide_layout_ = related_fixture
        slide_layout = slide_master_part.related_slide_layout(rId)
        getitem_.assert_called_once_with(rId)
        assert slide_layout is slide_layout_

    # fixtures -------------------------------------------------------

    @pytest.fixture
    def master_fixture(self, SlideMaster_, slide_master_):
        sldMaster = element("p:sldMaster")
        slide_master_part = SlideMasterPart(None, None, sldMaster)
        return slide_master_part, SlideMaster_, sldMaster, slide_master_

    @pytest.fixture
    def related_fixture(self, slide_layout_, related_parts_prop_):
        slide_master_part = SlideMasterPart(None, None, None, None)
        rId = "rId42"
        getitem_ = related_parts_prop_.return_value.__getitem__
        getitem_.return_value.slide_layout = slide_layout_
        return slide_master_part, rId, getitem_, slide_layout_

    # fixture components ---------------------------------------------

    @pytest.fixture
    def related_parts_prop_(self, request):
        return property_mock(request, SlideMasterPart, "related_parts")

    @pytest.fixture
    def slide_layout_(self, request):
        return instance_mock(request, SlideLayout)

    @pytest.fixture
    def SlideMaster_(self, request, slide_master_):
        return class_mock(
            request, "pptx.parts.slide.SlideMaster", return_value=slide_master_
        )

    @pytest.fixture
    def slide_master_(self, request):
        return instance_mock(request, SlideMaster)



# encoding: utf-8

"""
Test suite for pptx.oxml.ns.py module.
"""

from __future__ import print_function, unicode_literals

import pytest

from pptx.oxml.ns import NamespacePrefixedTag, namespaces, nsdecls, nsuri, qn


class DescribeNamespacePrefixedTag(object):
    def it_behaves_like_a_string_when_you_want_it_to(self, nsptag):
        s = "- %s -" % nsptag
        assert s == "- a:foobar -"

    def it_knows_its_clark_name(self, nsptag, clark_name):
        assert nsptag.clark_name == clark_name

    def it_knows_its_local_part(self, nsptag, local_part):
        assert nsptag.local_part == local_part

    def it_can_compose_a_single_entry_nsmap_for_itself(self, nsptag, namespace_uri_a):
        expected_nsmap = {"a": namespace_uri_a}
        assert nsptag.nsmap == expected_nsmap

    def it_knows_its_namespace_prefix(self, nsptag):
        assert nsptag.nspfx == "a"

    def it_knows_its_namespace_uri(self, nsptag, namespace_uri_a):
        assert nsptag.nsuri == namespace_uri_a


class DescribeNamespaces(object):
    def it_composes_a_dict_of_ns_uris_keyed_by_ns_pfx(self, nsmap):
        assert namespaces("a", "p") == nsmap


class DescribeNsdecls(object):
    def it_formats_namespace_declarations_from_a_list_of_prefixes(self, nsdecls_str):
        assert nsdecls("a", "p") == nsdecls_str


class DescribeNsuri(object):
    def it_finds_the_namespace_uri_corresponding_to_a_namespace_prefix(
        self, namespace_uri_a
    ):
        assert nsuri("a") == namespace_uri_a


class DescribeQn(object):
    def it_calculates_the_clark_name_for_an_ns_prefixed_tag_string(
        self, nsptag_str, clark_name
    ):
        assert qn(nsptag_str) == clark_name


# ===========================================================================
# fixtures
# ===========================================================================


@pytest.fixture
def clark_name(namespace_uri_a, local_part):
    return "{%s}%s" % (namespace_uri_a, local_part)


@pytest.fixture
def local_part():
    return "foobar"


@pytest.fixture
def namespace_uri_a():
    return "http://schemas.openxmlformats.org/drawingml/2006/main"


@pytest.fixture
def namespace_uri_p():
    return "http://schemas.openxmlformats.org/presentationml/2006/main"


@pytest.fixture
def nsdecls_str(namespace_uri_a, namespace_uri_p):
    return 'xmlns:a="%s" xmlns:p="%s"' % (namespace_uri_a, namespace_uri_p)


@pytest.fixture
def nsmap(namespace_uri_a, namespace_uri_p):
    return {"a": namespace_uri_a, "p": namespace_uri_p}


@pytest.fixture
def nsptag(nsptag_str):
    return NamespacePrefixedTag(nsptag_str)


@pytest.fixture
def nsptag_str(local_part):
    return "a:%s" % local_part



# encoding: utf-8

"""lxml custom element classes for DrawingML line-related XML elements."""

from __future__ import absolute_import, division, print_function, unicode_literals

from pptx.enum.dml import MSO_LINE_DASH_STYLE
from pptx.oxml.xmlchemy import BaseOxmlElement, OptionalAttribute


class CT_PresetLineDashProperties(BaseOxmlElement):
    """`a:prstDash` custom element class"""

    val = OptionalAttribute("val", MSO_LINE_DASH_STYLE)



# encoding: utf-8

"""Slide and related objects."""

from pptx.enum.shapes import PROG_ID
from pptx.opc.constants import CONTENT_TYPE as CT, RELATIONSHIP_TYPE as RT
from pptx.opc.package import XmlPart
from pptx.opc.packuri import PackURI
from pptx.oxml.slide import CT_NotesMaster, CT_NotesSlide, CT_Slide
from pptx.oxml.theme import CT_OfficeStyleSheet
from pptx.parts.chart import ChartPart
from pptx.parts.embeddedpackage import EmbeddedPackagePart
from pptx.slide import NotesMaster, NotesSlide, Slide, SlideLayout, SlideMaster
from pptx.util import lazyproperty


class BaseSlidePart(XmlPart):
    """Base class for slide parts.

    This includes slide, slide-layout, and slide-master parts, but also notes-slide,
    notes-master, and handout-master parts.
    """

    def get_image(self, rId):
        """
        Return an |Image| object containing the image related to this slide
        by *rId*. Raises |KeyError| if no image is related by that id, which
        would generally indicate a corrupted .pptx file.
        """
        return self.related_parts[rId].image

    def get_or_add_image_part(self, image_file):
        """
        Return an ``(image_part, rId)`` 2-tuple corresponding to an
        |ImagePart| object containing the image in *image_file*, and related
        to this slide with the key *rId*. If either the image part or
        relationship already exists, they are reused, otherwise they are
        newly created.
        """
        image_part = self._package.get_or_add_image_part(image_file)
        rId = self.relate_to(image_part, RT.IMAGE)
        return image_part, rId

    @property
    def name(self):
        """
        Internal name of this slide.
        """
        return self._element.cSld.name


class NotesMasterPart(BaseSlidePart):
    """Notes master part.

    Corresponds to package file `ppt/notesMasters/notesMaster1.xml`.
    """

    @classmethod
    def create_default(cls, package):
        """
        Create and return a default notes master part, including creating the
        new theme it requires.
        """
        notes_master_part = cls._new(package)
        theme_part = cls._new_theme_part(package)
        notes_master_part.relate_to(theme_part, RT.THEME)
        return notes_master_part

    @lazyproperty
    def notes_master(self):
        """
        Return the |NotesMaster| object that proxies this notes master part.
        """
        return NotesMaster(self._element, self)

    @classmethod
    def _new(cls, package):
        """
        Create and return a standalone, default notes master part based on
        the built-in template (without any related parts, such as theme).
        """
        partname = PackURI("/ppt/notesMasters/notesMaster1.xml")
        content_type = CT.PML_NOTES_MASTER
        notesMaster = CT_NotesMaster.new_default()
        return NotesMasterPart(partname, content_type, notesMaster, package)

    @classmethod
    def _new_theme_part(cls, package):
        """
        Create and return a default theme part suitable for use with a notes
        master.
        """
        partname = package.next_partname("/ppt/theme/theme%d.xml")
        content_type = CT.OFC_THEME
        theme = CT_OfficeStyleSheet.new_default()
        return XmlPart(partname, content_type, theme, package)


class NotesSlidePart(BaseSlidePart):
    """Notes slide part.

    Contains the slide notes content and the layout for the slide handout page.
    Corresponds to package file `ppt/notesSlides/notesSlide[1-9][0-9]*.xml`.
    """

    @classmethod
    def new(cls, package, slide_part):
        """
        Create and return a new notes slide part based on the notes master
        and related to both the notes master part and *slide_part*. If no
        notes master is present, create one based on the default template.
        """
        notes_master_part = package.presentation_part.notes_master_part
        notes_slide_part = cls._add_notes_slide_part(
            package, slide_part, notes_master_part
        )
        notes_slide = notes_slide_part.notes_slide
        notes_slide.clone_master_placeholders(notes_master_part.notes_master)
        return notes_slide_part

    @lazyproperty
    def notes_master(self):
        """
        Return the |NotesMaster| object this notes slide inherits from.
        """
        notes_master_part = self.part_related_by(RT.NOTES_MASTER)
        return notes_master_part.notes_master

    @lazyproperty
    def notes_slide(self):
        """
        Return the |NotesSlide| object that proxies this notes slide part.
        """
        return NotesSlide(self._element, self)

    @classmethod
    def _add_notes_slide_part(cls, package, slide_part, notes_master_part):
        """
        Create and return a new notes slide part that is fully related, but
        has no shape content (i.e. placeholders not cloned).
        """
        partname = package.next_partname("/ppt/notesSlides/notesSlide%d.xml")
        content_type = CT.PML_NOTES_SLIDE
        notes = CT_NotesSlide.new()
        notes_slide_part = NotesSlidePart(partname, content_type, notes, package)
        notes_slide_part.relate_to(notes_master_part, RT.NOTES_MASTER)
        notes_slide_part.relate_to(slide_part, RT.SLIDE)
        return notes_slide_part


class SlidePart(BaseSlidePart):
    """Slide part. Corresponds to package files ppt/slides/slide[1-9][0-9]*.xml."""

    @classmethod
    def new(cls, partname, package, slide_layout_part):
        """
        Return a newly-created blank slide part having *partname* and related
        to *slide_layout_part*.
        """
        sld = CT_Slide.new()
        slide_part = cls(partname, CT.PML_SLIDE, sld, package)
        slide_part.relate_to(slide_layout_part, RT.SLIDE_LAYOUT)
        return slide_part

    def add_chart_part(self, chart_type, chart_data):
        """
        Return the rId of a new |ChartPart| object containing a chart of
        *chart_type*, displaying *chart_data*, and related to the slide
        contained in this part.
        """
        chart_part = ChartPart.new(chart_type, chart_data, self.package)
        rId = self.relate_to(chart_part, RT.CHART)
        return rId

    def add_embedded_ole_object_part(self, prog_id, ole_object_file):
        """Return rId of newly-added OLE-object part formed from `ole_object_file`."""
        relationship_type = RT.PACKAGE if prog_id in PROG_ID else RT.OLE_OBJECT
        return self.relate_to(
            EmbeddedPackagePart.factory(
                prog_id, self._blob_from_file(ole_object_file), self._package
            ),
            relationship_type,
        )

    def get_or_add_video_media_part(self, video):
        """Return rIds for media and video relationships to media part.

        A new |MediaPart| object is created if it does not already exist
        (such as would occur if the same video appeared more than once in
         a presentation). Two relationships to the media part are created,
        one each with MEDIA and VIDEO relationship types. The need for two
        appears to be for legacy support for an earlier (pre-Office 2010)
        PowerPoint media embedding strategy.
        """
        media_part = self._package.get_or_add_media_part(video)
        media_rId = self.relate_to(media_part, RT.MEDIA)
        video_rId = self.relate_to(media_part, RT.VIDEO)
        return media_rId, video_rId

    @property
    def has_notes_slide(self):
        """
        Return True if this slide has a notes slide, False otherwise. A notes
        slide is created by the :attr:`notes_slide` property when one doesn't
        exist; use this property to test for a notes slide without the
        possible side-effect of creating one.
        """
        try:
            self.part_related_by(RT.NOTES_SLIDE)
        except KeyError:
            return False
        return True

    @lazyproperty
    def notes_slide(self):
        """
        The |NotesSlide| instance associated with this slide. If the slide
        does not have a notes slide, a new one is created. The same single
        instance is returned on each call.
        """
        try:
            notes_slide_part = self.part_related_by(RT.NOTES_SLIDE)
        except KeyError:
            notes_slide_part = self._add_notes_slide_part()
        return notes_slide_part.notes_slide

    @lazyproperty
    def slide(self):
        """
        The |Slide| object representing this slide part.
        """
        return Slide(self._element, self)

    @property
    def slide_id(self):
        """
        Return the slide identifier stored in the presentation part for this
        slide part.
        """
        presentation_part = self.package.presentation_part
        return presentation_part.slide_id(self)

    @property
    def slide_layout(self):
        """
        |SlideLayout| object the slide in this part inherits from.
        """
        slide_layout_part = self.part_related_by(RT.SLIDE_LAYOUT)
        return slide_layout_part.slide_layout

    def _add_notes_slide_part(self):
        """
        Return a newly created |NotesSlidePart| object related to this slide
        part. Caller is responsible for ensuring this slide doesn't already
        have a notes slide part.
        """
        notes_slide_part = NotesSlidePart.new(self.package, self)
        self.relate_to(notes_slide_part, RT.NOTES_SLIDE)
        return notes_slide_part


class SlideLayoutPart(BaseSlidePart):
    """Slide layout part.

    Corresponds to package files ``ppt/slideLayouts/slideLayout[1-9][0-9]*.xml``.
    """

    @lazyproperty
    def slide_layout(self):
        """
        The |SlideLayout| object representing this part.
        """
        return SlideLayout(self._element, self)

    @property
    def slide_master(self):
        """
        Slide master from which this slide layout inherits properties.
        """
        return self.part_related_by(RT.SLIDE_MASTER).slide_master


class SlideMasterPart(BaseSlidePart):
    """Slide master part.

    Corresponds to package files ppt/slideMasters/slideMaster[1-9][0-9]*.xml.
    """

    def related_slide_layout(self, rId):
        """
        Return the |SlideLayout| object of the related |SlideLayoutPart|
        corresponding to relationship key *rId*.
        """
        return self.related_parts[rId].slide_layout

    @lazyproperty
    def slide_master(self):
        """
        The |SlideMaster| object representing this part.
        """
        return SlideMaster(self._element, self)



# encoding: utf-8

"""
Test suite for pptx.chart.xmlwriter module
"""

from __future__ import absolute_import, print_function, unicode_literals

from datetime import date
from itertools import islice

import pytest

from pptx.chart.data import (
    _BaseChartData,
    _BaseSeriesData,
    BubbleChartData,
    CategoryChartData,
    CategorySeriesData,
    XyChartData,
)
from pptx.chart.xmlwriter import (
    _AreaChartXmlWriter,
    _BarChartXmlWriter,
    _BaseSeriesXmlRewriter,
    _BubbleChartXmlWriter,
    _BubbleSeriesXmlRewriter,
    _BubbleSeriesXmlWriter,
    _CategorySeriesXmlRewriter,
    _CategorySeriesXmlWriter,
    ChartXmlWriter,
    _DoughnutChartXmlWriter,
    _LineChartXmlWriter,
    _PieChartXmlWriter,
    _RadarChartXmlWriter,
    SeriesXmlRewriterFactory,
    _XyChartXmlWriter,
    _XySeriesXmlRewriter,
    _XySeriesXmlWriter,
)
from pptx.enum.chart import XL_CHART_TYPE
from pptx.oxml import parse_xml

from ..unitutil import count
from ..unitutil.cxml import element, xml
from ..unitutil.file import snippet_seq, snippet_text
from ..unitutil.mock import call, class_mock, instance_mock, method_mock


class DescribeChartXmlWriter(object):
    def it_contructs_an_xml_writer_for_a_chart_type(self, call_fixture):
        chart_type, series_seq_, XmlWriterClass_, xml_writer_ = call_fixture
        xml_writer = ChartXmlWriter(chart_type, series_seq_)
        XmlWriterClass_.assert_called_once_with(chart_type, series_seq_)
        assert xml_writer is xml_writer_

    # fixtures -------------------------------------------------------

    @pytest.fixture(
        params=[
            ("AREA", _AreaChartXmlWriter),
            ("AREA_STACKED", _AreaChartXmlWriter),
            ("AREA_STACKED_100", _AreaChartXmlWriter),
            ("BAR_CLUSTERED", _BarChartXmlWriter),
            ("BAR_STACKED", _BarChartXmlWriter),
            ("BAR_STACKED_100", _BarChartXmlWriter),
            ("BUBBLE", _BubbleChartXmlWriter),
            ("BUBBLE_THREE_D_EFFECT", _BubbleChartXmlWriter),
            ("COLUMN_CLUSTERED", _BarChartXmlWriter),
            ("COLUMN_STACKED", _BarChartXmlWriter),
            ("COLUMN_STACKED_100", _BarChartXmlWriter),
            ("DOUGHNUT", _DoughnutChartXmlWriter),
            ("DOUGHNUT_EXPLODED", _DoughnutChartXmlWriter),
            ("LINE", _LineChartXmlWriter),
            ("LINE_MARKERS", _LineChartXmlWriter),
            ("LINE_MARKERS_STACKED", _LineChartXmlWriter),
            ("LINE_MARKERS_STACKED_100", _LineChartXmlWriter),
            ("LINE_STACKED", _LineChartXmlWriter),
            ("LINE_STACKED_100", _LineChartXmlWriter),
            ("PIE", _PieChartXmlWriter),
            ("PIE_EXPLODED", _PieChartXmlWriter),
            ("RADAR", _RadarChartXmlWriter),
            ("RADAR_FILLED", _RadarChartXmlWriter),
            ("RADAR_MARKERS", _RadarChartXmlWriter),
            ("XY_SCATTER", _XyChartXmlWriter),
            ("XY_SCATTER_LINES", _XyChartXmlWriter),
            ("XY_SCATTER_LINES_NO_MARKERS", _XyChartXmlWriter),
            ("XY_SCATTER_SMOOTH", _XyChartXmlWriter),
            ("XY_SCATTER_SMOOTH_NO_MARKERS", _XyChartXmlWriter),
        ]
    )
    def call_fixture(self, request, series_seq_):
        chart_type_member, XmlWriterClass = request.param
        xml_writer_ = instance_mock(request, XmlWriterClass)
        class_spec = "pptx.chart.xmlwriter.%s" % XmlWriterClass.__name__
        XmlWriterClass_ = class_mock(request, class_spec, return_value=xml_writer_)
        chart_type = getattr(XL_CHART_TYPE, chart_type_member)
        return chart_type, series_seq_, XmlWriterClass_, xml_writer_

    # fixture components ---------------------------------------------

    @pytest.fixture
    def series_seq_(self, request):
        return instance_mock(request, tuple)


class DescribeSeriesXmlRewriterFactory(object):
    def it_contructs_an_xml_rewriter_for_a_chart_type(self, call_fixture):
        chart_type, chart_data_, XmlRewriterClass_, xml_rewriter_ = call_fixture

        xml_rewriter = SeriesXmlRewriterFactory(chart_type, chart_data_)

        XmlRewriterClass_.assert_called_once_with(chart_data_)
        assert xml_rewriter is xml_rewriter_

    # fixtures -------------------------------------------------------

    @pytest.fixture(
        params=[
            ("BAR_CLUSTERED", _CategorySeriesXmlRewriter),
            ("BUBBLE", _BubbleSeriesXmlRewriter),
            ("BUBBLE_THREE_D_EFFECT", _BubbleSeriesXmlRewriter),
            ("XY_SCATTER", _XySeriesXmlRewriter),
            ("XY_SCATTER_LINES", _XySeriesXmlRewriter),
            ("XY_SCATTER_LINES_NO_MARKERS", _XySeriesXmlRewriter),
            ("XY_SCATTER_SMOOTH", _XySeriesXmlRewriter),
            ("XY_SCATTER_SMOOTH_NO_MARKERS", _XySeriesXmlRewriter),
        ]
    )
    def call_fixture(self, request, chart_data_):
        chart_type_member, rewriter_cls = request.param
        chart_type = getattr(XL_CHART_TYPE, chart_type_member)
        xml_rewriter_ = instance_mock(request, rewriter_cls)
        XmlRewriterClass_ = class_mock(
            request,
            "pptx.chart.xmlwriter.%s" % rewriter_cls.__name__,
            return_value=xml_rewriter_,
        )
        return chart_type, chart_data_, XmlRewriterClass_, xml_rewriter_

    # fixture components ---------------------------------------------

    @pytest.fixture
    def chart_data_(self, request):
        return instance_mock(request, _BaseChartData)


class Describe_AreaChartXmlWriter(object):
    def it_can_generate_xml_for_area_type_charts(self, xml_fixture):
        xml_writer, expected_xml = xml_fixture
        assert xml_writer.xml == expected_xml

    # fixtures -------------------------------------------------------

    @pytest.fixture(
        params=[
            ("AREA", 2, 2, str, "2x2-area"),
            ("AREA", 2, 2, date, "2x2-area-date"),
            ("AREA", 2, 2, float, "2x2-area-float"),
            ("AREA_STACKED", 2, 2, str, "2x2-area-stacked"),
            ("AREA_STACKED_100", 2, 2, str, "2x2-area-stacked-100"),
        ]
    )
    def xml_fixture(self, request):
        member, cat_count, ser_count, cat_type, snippet_name = request.param
        chart_type = getattr(XL_CHART_TYPE, member)
        chart_data = make_category_chart_data(cat_count, cat_type, ser_count)
        xml_writer = _AreaChartXmlWriter(chart_type, chart_data)
        expected_xml = snippet_text(snippet_name)
        return xml_writer, expected_xml


class Describe_BarChartXmlWriter(object):
    """Unit-test suite for `pptx.chart.xmlwriter._BarChartXmlWriter`."""

    @pytest.mark.parametrize(
        "member, cat_count, ser_count, cat_type, snippet_name",
        (
            ("BAR_CLUSTERED", 2, 2, str, "2x2-bar-clustered"),
            ("BAR_CLUSTERED", 2, 2, date, "2x2-bar-clustered-date"),
            ("BAR_CLUSTERED", 2, 2, float, "2x2-bar-clustered-float"),
            ("BAR_STACKED", 2, 2, str, "2x2-bar-stacked"),
            ("BAR_STACKED_100", 2, 2, str, "2x2-bar-stacked-100"),
            ("COLUMN_CLUSTERED", 2, 2, str, "2x2-column-clustered"),
            ("COLUMN_STACKED", 2, 2, str, "2x2-column-stacked"),
            ("COLUMN_STACKED_100", 2, 2, str, "2x2-column-stacked-100"),
        ),
    )
    def it_can_generate_xml_for_bar_type_charts(
        self, member, cat_count, ser_count, cat_type, snippet_name
    ):
        chart_type = getattr(XL_CHART_TYPE, member)
        chart_data = make_category_chart_data(cat_count, cat_type, ser_count)
        xml_writer = _BarChartXmlWriter(chart_type, chart_data)

        assert xml_writer.xml == snippet_text(snippet_name)

    def it_can_generate_xml_for_multi_level_cat_charts(self):
        chart_data = CategoryChartData()
        WEST = chart_data.add_category("WEST")
        WEST.add_sub_category("SF")
        WEST.add_sub_category("LA")
        EAST = chart_data.add_category("EAST")
        EAST.add_sub_category("NY")
        EAST.add_sub_category("NJ")
        chart_data.add_series("Series 1", (1, 2, None, 4))
        chart_data.add_series("Series 2", (5, None, 7, 8))
        xml_writer = _BarChartXmlWriter(XL_CHART_TYPE.BAR_CLUSTERED, chart_data)

        assert xml_writer.xml == snippet_text("4x2-multi-cat-bar")


class Describe_BubbleChartXmlWriter(object):
    def it_can_generate_xml_for_bubble_charts(self, xml_fixture):
        xml_writer, expected_xml = xml_fixture
        assert xml_writer.xml == expected_xml

    # fixtures -------------------------------------------------------

    @pytest.fixture(
        params=[
            ("BUBBLE", 2, 3, "2x3-bubble"),
            ("BUBBLE_THREE_D_EFFECT", 2, 3, "2x3-bubble-3d"),
        ]
    )
    def xml_fixture(self, request):
        enum_member, ser_count, point_count, snippet_name = request.param
        chart_type = getattr(XL_CHART_TYPE, enum_member)
        chart_data = make_bubble_chart_data(ser_count, point_count)
        xml_writer = _BubbleChartXmlWriter(chart_type, chart_data)
        expected_xml = snippet_text(snippet_name)
        return xml_writer, expected_xml


class Describe_DoughnutChartXmlWriter(object):
    def it_can_generate_xml_for_doughnut_type_charts(self, xml_fixture):
        xml_writer, expected_xml = xml_fixture
        assert xml_writer.xml == expected_xml

    # fixtures -------------------------------------------------------

    @pytest.fixture(
        params=[
            ("DOUGHNUT", 3, 2, "3x2-doughnut"),
            ("DOUGHNUT_EXPLODED", 3, 2, "3x2-doughnut-exploded"),
        ]
    )
    def xml_fixture(self, request):
        enum_member, cat_count, ser_count, snippet_name = request.param
        chart_type = getattr(XL_CHART_TYPE, enum_member)
        chart_data = make_category_chart_data(cat_count, str, ser_count)
        xml_writer = _DoughnutChartXmlWriter(chart_type, chart_data)
        expected_xml = snippet_text(snippet_name)
        return xml_writer, expected_xml


class Describe_LineChartXmlWriter(object):
    def it_can_generate_xml_for_a_line_chart(self, xml_fixture):
        xml_writer, expected_xml = xml_fixture
        assert xml_writer.xml == expected_xml

    # fixtures -------------------------------------------------------

    @pytest.fixture(
        params=[
            ("LINE", 2, 2, str, "2x2-line"),
            ("LINE", 2, 2, date, "2x2-line-date"),
            ("LINE", 2, 2, float, "2x2-line-float"),
            ("LINE_MARKERS", 2, 2, str, "2x2-line-markers"),
            ("LINE_MARKERS_STACKED", 2, 2, str, "2x2-line-markers-stacked"),
            ("LINE_MARKERS_STACKED_100", 2, 2, str, "2x2-line-markers-stacked-100"),
            ("LINE_STACKED", 2, 2, str, "2x2-line-stacked"),
            ("LINE_STACKED_100", 2, 2, str, "2x2-line-stacked-100"),
        ]
    )
    def xml_fixture(self, request):
        member, cat_count, ser_count, cat_type, snippet_name = request.param
        chart_type = getattr(XL_CHART_TYPE, member)
        chart_data = make_category_chart_data(cat_count, cat_type, ser_count)
        xml_writer = _LineChartXmlWriter(chart_type, chart_data)
        expected_xml = snippet_text(snippet_name)
        return xml_writer, expected_xml


class Describe_PieChartXmlWriter(object):
    """Unit-test suite for `pptx.chart.xmlwriter._PieChartXmlWriter`."""

    @pytest.mark.parametrize(
        "enum_member, cat_count, ser_count, snippet_name",
        (
            ("PIE", 3, 1, "3x1-pie"),
            ("PIE_EXPLODED", 3, 1, "3x1-pie-exploded"),
        ),
    )
    def it_can_generate_xml_for_a_pie_chart(
        self, enum_member, cat_count, ser_count, snippet_name
    ):
        chart_type = getattr(XL_CHART_TYPE, enum_member)
        chart_data = make_category_chart_data(cat_count, str, ser_count)
        xml_writer = _PieChartXmlWriter(chart_type, chart_data)

        assert xml_writer.xml == snippet_text(snippet_name)


class Describe_RadarChartXmlWriter(object):
    """Unit-test suite for `pptx.chart.xmlwriter._RadarChartXmlWriter`."""

    def it_can_generate_xml_for_a_radar_chart(self):
        series_data_seq = make_category_chart_data(
            cat_count=5, cat_type=str, ser_count=2
        )
        xml_writer = _RadarChartXmlWriter(XL_CHART_TYPE.RADAR, series_data_seq)

        assert xml_writer.xml == snippet_text("2x5-radar")


class Describe_XyChartXmlWriter(object):
    def it_can_generate_xml_for_xy_charts(self, xml_fixture):
        xml_writer, expected_xml = xml_fixture
        assert xml_writer.xml == expected_xml

    # fixtures -------------------------------------------------------

    @pytest.fixture(
        params=[
            ("XY_SCATTER", 2, 3, "2x3-xy"),
            ("XY_SCATTER_LINES", 2, 3, "2x3-xy-lines"),
            ("XY_SCATTER_LINES_NO_MARKERS", 2, 3, "2x3-xy-lines-no-markers"),
            ("XY_SCATTER_SMOOTH", 2, 3, "2x3-xy-smooth"),
            ("XY_SCATTER_SMOOTH_NO_MARKERS", 2, 3, "2x3-xy-smooth-no-markers"),
        ]
    )
    def xml_fixture(self, request):
        enum_member, ser_count, point_count, snippet_name = request.param
        chart_type = getattr(XL_CHART_TYPE, enum_member)
        chart_data = make_xy_chart_data(ser_count, point_count)
        xml_writer = _XyChartXmlWriter(chart_type, chart_data)
        expected_xml = snippet_text(snippet_name)
        return xml_writer, expected_xml


class Describe_BubbleSeriesXmlWriter(object):
    def it_knows_its_bubbleSize_XML(self, bubbleSize_fixture):
        xml_writer, expected_xml = bubbleSize_fixture
        bubbleSize = xml_writer.bubbleSize
        assert bubbleSize.xml == expected_xml

    # fixtures -------------------------------------------------------

    @pytest.fixture(
        params=[
            (None, None, "General"),
            (None, 42, "42"),
            (24, None, "24"),
            (24, 42, "42"),
        ]
    )
    def bubbleSize_fixture(self, request):
        cht_numfmt, ser_numfmt, expected_numfmt = request.param
        expected_xml = xml(
            'c:bubbleSize/c:numRef/(c:f"Sheet1!$C$2:$C$1",c:numCache/(c:form'
            'atCode"%s",c:ptCount{val=0}))' % expected_numfmt
        )
        chart_number_format = () if cht_numfmt is None else (cht_numfmt,)
        series_number_format = () if ser_numfmt is None else (ser_numfmt,)
        chart_data = BubbleChartData(*chart_number_format)
        series_data = chart_data.add_series(None, *series_number_format)
        xml_writer = _BubbleSeriesXmlWriter(series_data)
        return xml_writer, expected_xml


class Describe_CategorySeriesXmlWriter(object):
    def it_knows_its_val_XML(self, val_fixture):
        xml_writer, expected_xml = val_fixture
        val = xml_writer.val
        assert val.xml == expected_xml

    # fixtures -------------------------------------------------------

    @pytest.fixture
    def val_fixture(self, series_data_):
        values_ref, number_format = "Sheet1!$B$2:$B$1", "Foobar"
        expected_xml = xml(
            'c:val/c:numRef/(c:f"%s",c:numCache/(c:formatCode"%s",c:ptCount{'
            "val=0}))" % (values_ref, number_format)
        )
        xml_writer = _CategorySeriesXmlWriter(series_data_)
        series_data_.values_ref = values_ref
        series_data_.number_format = number_format
        return xml_writer, expected_xml

    # fixture components ---------------------------------------------

    @pytest.fixture
    def series_data_(self, request):
        return instance_mock(request, CategorySeriesData)


class Describe_XySeriesXmlWriter(object):
    def it_knows_its_xVal_XML(self, xVal_fixture):
        xml_writer, expected_xml = xVal_fixture
        xVal = xml_writer.xVal
        assert xVal.xml == expected_xml

    def it_knows_its_yVal_XML(self, yVal_fixture):
        xml_writer, expected_xml = yVal_fixture
        yVal = xml_writer.yVal
        assert yVal.xml == expected_xml

    # fixtures -------------------------------------------------------

    @pytest.fixture(
        params=[
            (None, None, "General"),
            (None, 42, "42"),
            (24, None, "24"),
            (24, 42, "42"),
        ]
    )
    def xVal_fixture(self, request):
        cht_numfmt, ser_numfmt, expected_numfmt = request.param
        expected_xml = xml(
            'c:xVal/c:numRef/(c:f"Sheet1!$A$2:$A$1",c:numCache/(c:formatCode'
            '"%s",c:ptCount{val=0}))' % expected_numfmt
        )
        chart_number_format = () if cht_numfmt is None else (cht_numfmt,)
        series_number_format = () if ser_numfmt is None else (ser_numfmt,)
        chart_data = XyChartData(*chart_number_format)
        series_data = chart_data.add_series(None, *series_number_format)
        xml_writer = _XySeriesXmlWriter(series_data)
        return xml_writer, expected_xml

    @pytest.fixture(
        params=[
            (None, None, "General"),
            (None, 42, "42"),
            (24, None, "24"),
            (24, 42, "42"),
        ]
    )
    def yVal_fixture(self, request):
        cht_numfmt, ser_numfmt, expected_numfmt = request.param
        expected_xml = xml(
            'c:yVal/c:numRef/(c:f"Sheet1!$B$2:$B$1",c:numCache/(c:formatCode'
            '"%s",c:ptCount{val=0}))' % expected_numfmt
        )
        chart_number_format = () if cht_numfmt is None else (cht_numfmt,)
        series_number_format = () if ser_numfmt is None else (ser_numfmt,)
        chart_data = XyChartData(*chart_number_format)
        series_data = chart_data.add_series(None, *series_number_format)
        xml_writer = _XySeriesXmlWriter(series_data)
        return xml_writer, expected_xml


class Describe_BaseSeriesXmlRewriter(object):
    def it_can_replace_series_data(self, replace_fixture):
        rewriter, chartSpace, plotArea, ser_count, calls = replace_fixture
        rewriter.replace_series_data(chartSpace)
        rewriter._adjust_ser_count.assert_called_once_with(
            rewriter, plotArea, ser_count
        )
        assert rewriter._rewrite_ser_data.call_args_list == calls

    def it_adjusts_the_ser_count_to_help(self, adjust_fixture):
        rewriter, plotArea, new_ser_count, add_calls, trim_calls, sers = adjust_fixture
        rewriter._adjust_ser_count(plotArea, new_ser_count)
        assert rewriter._add_cloned_sers.call_args_list == add_calls
        assert rewriter._trim_ser_count_by.call_args_list == trim_calls

    def it_adds_cloned_sers_to_help(self, clone_fixture):
        rewriter, plotArea, count, expected_xml = clone_fixture
        rewriter._add_cloned_sers(plotArea, count)
        assert plotArea.xml == expected_xml

    def it_trims_sers_to_help(self, trim_fixture):
        rewriter, plotArea, count, expected_xml = trim_fixture
        rewriter._trim_ser_count_by(plotArea, count)
        assert plotArea.xml == expected_xml

    # fixtures -------------------------------------------------------

    @pytest.fixture(params=[(3, True, False), (2, False, False), (1, False, True)])
    def adjust_fixture(self, request, _add_cloned_sers_, _trim_ser_count_by_):
        new_ser_count, add, trim = request.param
        rewriter = _BaseSeriesXmlRewriter(None)
        plotArea = element(
            "c:plotArea/c:barChart/(c:ser/(c:idx{val=3},c:order{val=1}),c:se"
            "r/(c:idx{val=1},c:order{val=3}))"
        )
        sers = plotArea.sers
        add_calls = [call(rewriter, plotArea, 1)] if add else []
        trim_calls = [call(rewriter, plotArea, 1)] if trim else []
        return rewriter, plotArea, new_ser_count, add_calls, trim_calls, sers

    @pytest.fixture(
        params=[
            (
                "c:plotArea/c:barChart/c:ser/(c:idx{val=3},c:order{val=1})",
                1,
                "c:plotArea/c:barChart/(c:ser/(c:idx{val=3},c:order{val=1}),c:ser/("
                "c:idx{val=4},c:order{val=2}))",
            ),
            (
                "c:plotArea/(c:barChart/c:ser/(c:idx{val=3},c:order{val=1}),c:lineC"
                "hart/(c:ser/(c:idx{val=4},c:order{val=7}),c:ser/(c:idx{val=1},c:or"
                "der{val=2})))",
                1,
                "c:plotArea/(c:barChart/c:ser/(c:idx{val=3},c:order{val=1}),c:lineC"
                "hart/(c:ser/(c:idx{val=4},c:order{val=7}),c:ser/(c:idx{val=5},c:or"
                "der{val=8}),c:ser/(c:idx{val=1},c:order{val=2})))",
            ),
        ]
    )
    def clone_fixture(self, request):
        plotArea_cxml, count, expected_cxml = request.param
        rewriter = _BaseSeriesXmlRewriter(None)
        plotArea = element(plotArea_cxml)
        expected_xml = xml(expected_cxml)
        return rewriter, plotArea, count, expected_xml

    @pytest.fixture
    def replace_fixture(
        self, request, chart_data_, _adjust_ser_count_, _rewrite_ser_data_
    ):
        rewriter = _BaseSeriesXmlRewriter(chart_data_)
        chartSpace = element(
            "c:chartSpace/c:chart/c:plotArea/c:barChart/(c:ser/c:order{val=0"
            "},c:ser/c:order{val=1})"
        )
        plotArea = chartSpace.xpath("c:chart/c:plotArea")[0]
        sers = chartSpace.xpath(".//c:ser")
        ser_count = len(sers)
        series_datas = [
            instance_mock(request, _BaseSeriesData),
            instance_mock(request, _BaseSeriesData),
        ]
        calls = [
            call(rewriter, sers[0], series_datas[0], False),
            call(rewriter, sers[1], series_datas[1], False),
        ]
        chart_data_.__len__.return_value = len(series_datas)
        chart_data_.__iter__.return_value = iter(series_datas)
        return rewriter, chartSpace, plotArea, ser_count, calls

    @pytest.fixture(
        params=[
            # --single xChart, remove single ser--
            (
                "c:plotArea/c:barChart/(c:ser/(c:idx{val=3},c:order{val=4}),c:ser/("
                "c:idx{val=1},c:order{val=2}))",
                1,
                "c:plotArea/c:barChart/(c:ser/(c:idx{val=1},c:order{val=2}))",
            ),
            # --two xCharts, remove two sers (and emptied second xChart)--
            (
                "c:plotArea/(c:barChart/c:ser/(c:idx{val=3},c:order{val=1}),c:lineC"
                "hart/(c:ser/(c:idx{val=4},c:order{val=7}),c:ser/(c:idx{val=1},c:or"
                "der{val=2})))",
                2,
                "c:plotArea/(c:barChart/c:ser/(c:idx{val=3},c:order{val=1}))",
            ),
        ]
    )
    def trim_fixture(self, request):
        plotArea_cxml, count, expected_cxml = request.param
        rewriter = _BaseSeriesXmlRewriter(None)
        plotArea = element(plotArea_cxml)
        expected_xml = xml(expected_cxml)
        return rewriter, plotArea, count, expected_xml

    # fixture components ---------------------------------------------

    @pytest.fixture
    def _add_cloned_sers_(self, request):
        return method_mock(
            request, _BaseSeriesXmlRewriter, "_add_cloned_sers", autospec=True
        )

    @pytest.fixture
    def _adjust_ser_count_(self, request):
        return method_mock(
            request, _BaseSeriesXmlRewriter, "_adjust_ser_count", autospec=True
        )

    @pytest.fixture
    def chart_data_(self, request):
        return instance_mock(request, _BaseChartData)

    @pytest.fixture
    def _rewrite_ser_data_(self, request):
        return method_mock(
            request, _BaseSeriesXmlRewriter, "_rewrite_ser_data", autospec=True
        )

    @pytest.fixture
    def _trim_ser_count_by_(self, request):
        return method_mock(
            request, _BaseSeriesXmlRewriter, "_trim_ser_count_by", autospec=True
        )


class Describe_BubbleSeriesXmlRewriter(object):
    def it_can_rewrite_a_ser_element(self, rewrite_fixture):
        rewriter, ser, series_data, expected_xml = rewrite_fixture
        rewriter._rewrite_ser_data(ser, series_data, None)
        assert ser.xml == expected_xml

    # fixtures -------------------------------------------------------

    @pytest.fixture
    def rewrite_fixture(self):
        ser_xml, expected_xml = snippet_seq("rewrite-ser")[4:6]

        chart_data = BubbleChartData()
        series_data = chart_data.add_series("Series 1")
        series_data.add_data_point(1, 2, 10)
        series_data.add_data_point(3, 4, 20)

        rewriter = _BubbleSeriesXmlRewriter(chart_data)
        ser = parse_xml(ser_xml)
        return rewriter, ser, series_data, expected_xml


class Describe_CategorySeriesXmlRewriter(object):
    def it_can_rewrite_a_ser_element(self, rewrite_fixture):
        rewriter, ser, series_data, expected_xml = rewrite_fixture
        rewriter._rewrite_ser_data(ser, series_data, False)
        assert ser.xml == expected_xml

    # fixtures -------------------------------------------------------

    @pytest.fixture(
        params=[
            (0, ("foo", "bar")),
            (6, (42, 24)),
            (8, (date(2016, 12, 22), date(2016, 12, 23))),
        ]
    )
    def rewrite_fixture(self, request):
        snippet_offset, categories = request.param
        chart_data = CategoryChartData()
        chart_data.categories = categories
        rewriter = _CategorySeriesXmlRewriter(chart_data)
        snippets = snippet_seq("rewrite-ser")
        ser_xml = snippets[snippet_offset]
        ser = parse_xml(ser_xml)
        series_data = chart_data.add_series("Series 1", (1, 2))
        expected_xml = snippets[snippet_offset + 1]
        return rewriter, ser, series_data, expected_xml


class Describe_XySeriesXmlRewriter(object):
    def it_can_rewrite_a_ser_element(self, rewrite_fixture):
        rewriter, ser, series_data, expected_xml = rewrite_fixture
        rewriter._rewrite_ser_data(ser, series_data, None)
        assert ser.xml == expected_xml

    # fixtures -------------------------------------------------------

    @pytest.fixture
    def rewrite_fixture(self):
        ser_xml, expected_xml = snippet_seq("rewrite-ser")[2:4]

        chart_data = XyChartData()
        series_data = chart_data.add_series("Series 1")
        series_data.add_data_point(1, 2)
        series_data.add_data_point(3, 4)

        rewriter = _XySeriesXmlRewriter(chart_data)
        ser = parse_xml(ser_xml)
        return rewriter, ser, series_data, expected_xml


# helpers ------------------------------------------------------------


def make_bubble_chart_data(ser_count, point_count):
    """
    Return an |BubbleChartData| object populated with *ser_count* series,
    each having *point_count* data points.
    """
    points = (
        (1.1, 11.1, 10.0),
        (2.1, 12.1, 20.0),
        (3.1, 13.1, 30.0),
        (1.2, 11.2, 40.0),
        (2.2, 12.2, 50.0),
        (3.2, 13.2, 60.0),
    )
    chart_data = BubbleChartData()
    for i in range(ser_count):
        series_label = "Series %d" % (i + 1)
        series = chart_data.add_series(series_label)
        for j in range(point_count):
            point_idx = (i * point_count) + j
            x, y, size = points[point_idx]
            series.add_data_point(x, y, size)
    return chart_data


def make_category_chart_data(cat_count, cat_type, ser_count):
    """
    Return a |CategoryChartData| instance populated with *cat_count*
    categories of type *cat_type* and *ser_count* series. Values are
    auto-generated.
    """
    category_labels = {
        date: (date(2016, 12, 27), date(2016, 12, 28), date(2016, 12, 29)),
        float: (1.1, 2.2, 3.3, 4.4, 5.5),
        str: ("Foo", "Bar", "Baz", "Boo", "Far", "Faz"),
    }[cat_type]
    point_values = count(1.1, 1.1)
    chart_data = CategoryChartData()
    chart_data.categories = category_labels[:cat_count]
    for idx in range(ser_count):
        series_title = "Series %d" % (idx + 1)
        series_values = tuple(islice(point_values, cat_count))
        series_values = [round(x * 10) / 10.0 for x in series_values]
        chart_data.add_series(series_title, series_values)
    return chart_data


def make_xy_chart_data(ser_count, point_count):
    """
    Return an |XyChartData| object populated with *ser_count* series each
    having *point_count* data points. Values are auto-generated.
    """
    points = (
        (1.1, 11.1),
        (2.1, 12.1),
        (3.1, 13.1),
        (1.2, 11.2),
        (2.2, 12.2),
        (3.2, 13.2),
    )
    chart_data = XyChartData()
    for i in range(ser_count):
        series_label = "Series %d" % (i + 1)
        series = chart_data.add_series(series_label)
        for j in range(point_count):
            point_idx = (i * point_count) + j
            x, y = points[point_idx]
            series.add_data_point(x, y)
    return chart_data



# encoding: utf-8

"""
Test suite for pptx.dml.line module
"""

from __future__ import absolute_import, print_function, unicode_literals

import pytest

from pptx.dml.color import ColorFormat
from pptx.dml.fill import FillFormat
from pptx.dml.line import LineFormat
from pptx.enum.dml import MSO_FILL, MSO_LINE
from pptx.oxml.shapes.shared import CT_LineProperties
from pptx.shapes.autoshape import Shape

from ..oxml.unitdata.dml import an_ln
from ..unitutil.cxml import element, xml
from ..unitutil.mock import call, class_mock, instance_mock, property_mock


class DescribeLineFormat(object):
    def it_knows_its_dash_style(self, dash_style_get_fixture):
        line, expected_value = dash_style_get_fixture
        assert line.dash_style == expected_value

    def it_can_change_its_dash_style(self, dash_style_set_fixture):
        line, dash_style, spPr, expected_xml = dash_style_set_fixture
        line.dash_style = dash_style
        assert spPr.xml == expected_xml

    def it_knows_its_width(self, width_get_fixture):
        line, expected_line_width = width_get_fixture
        assert line.width == expected_line_width

    def it_can_change_its_width(self, width_set_fixture):
        line, width, expected_xml = width_set_fixture
        line.width = width
        assert line._ln.xml == expected_xml

    def it_has_a_fill(self, fill_fixture):
        line, FillFormat_, ln_, fill_ = fill_fixture
        fill = line.fill
        FillFormat_.from_fill_parent.assert_called_once_with(ln_)
        assert fill is fill_

    def it_has_a_color(self, color_fixture):
        line, fill_, expected_solid_calls, color_ = color_fixture
        color = line.color
        assert fill_.solid.mock_calls == expected_solid_calls
        assert color is color_

    # fixtures -------------------------------------------------------

    @pytest.fixture(
        params=[(MSO_FILL.SOLID, False), (MSO_FILL.BACKGROUND, True), (None, True)]
    )
    def color_fixture(self, request, line, fill_prop_, fill_, color_):
        pre_call_fill_type, solid_call_expected = request.param
        fill_.type = pre_call_fill_type
        expected_solid_calls = [call()] if solid_call_expected else []
        return line, fill_, expected_solid_calls, color_

    @pytest.fixture(
        params=[
            ("p:spPr", None),
            ("p:spPr/a:ln", None),
            ("p:spPr/a:ln/a:prstDash", None),
            ("p:spPr/a:ln/a:prstDash{val=dash}", MSO_LINE.DASH),
            ("p:spPr/a:ln/a:prstDash{val=solid}", MSO_LINE.SOLID),
        ]
    )
    def dash_style_get_fixture(self, request):
        spPr_cxml, expected_value = request.param
        spPr = element(spPr_cxml)
        line = LineFormat(spPr)
        return line, expected_value

    @pytest.fixture(
        params=[
            ("p:spPr{a:b=c}", MSO_LINE.DASH, "p:spPr{a:b=c}/a:ln/a:prstDash{val=dash}"),
            ("p:spPr/a:ln", MSO_LINE.ROUND_DOT, "p:spPr/a:ln/a:prstDash{val=dot}"),
            (
                "p:spPr/a:ln/a:prstDash",
                MSO_LINE.SOLID,
                "p:spPr/a:ln/a:prstDash{val=solid}",
            ),
            (
                "p:spPr/a:ln/a:custDash",
                MSO_LINE.DASH_DOT,
                "p:spPr/a:ln/a:prstDash{val=dashDot}",
            ),
            (
                "p:spPr/a:ln/a:prstDash{val=dash}",
                MSO_LINE.LONG_DASH,
                "p:spPr/a:ln/a:prstDash{val=lgDash}",
            ),
            ("p:spPr/a:ln/a:prstDash{val=dash}", None, "p:spPr/a:ln"),
            ("p:spPr/a:ln/a:custDash", None, "p:spPr/a:ln"),
        ]
    )
    def dash_style_set_fixture(self, request):
        spPr_cxml, dash_style, expected_cxml = request.param
        spPr = element(spPr_cxml)
        line = LineFormat(spPr)
        expected_xml = xml(expected_cxml)
        return line, dash_style, spPr, expected_xml

    @pytest.fixture
    def fill_fixture(self, line, FillFormat_, ln_, fill_):
        return line, FillFormat_, ln_, fill_

    @pytest.fixture(params=[(None, 0), (12700, 12700)])
    def width_get_fixture(self, request, shape_):
        w, expected_line_width = request.param
        shape_.ln = self.ln_bldr(w).element
        line = LineFormat(shape_)
        return line, expected_line_width

    @pytest.fixture(
        params=[
            (None, None),
            (None, 12700),
            (12700, 12700),
            (12700, 25400),
            (25400, None),
        ]
    )
    def width_set_fixture(self, request, shape_):
        initial_width, width = request.param
        shape_.ln = shape_.get_or_add_ln.return_value = self.ln_bldr(
            initial_width
        ).element
        line = LineFormat(shape_)
        expected_xml = self.ln_bldr(width).xml()
        return line, width, expected_xml

    # fixture components ---------------------------------------------

    @pytest.fixture
    def color_(self, request):
        return instance_mock(request, ColorFormat)

    @pytest.fixture
    def fill_(self, request, color_):
        return instance_mock(request, FillFormat, fore_color=color_)

    @pytest.fixture
    def fill_prop_(self, request, fill_):
        return property_mock(request, LineFormat, "fill", return_value=fill_)

    @pytest.fixture
    def FillFormat_(self, request, fill_):
        FillFormat_ = class_mock(request, "pptx.dml.line.FillFormat")
        FillFormat_.from_fill_parent.return_value = fill_
        return FillFormat_

    @pytest.fixture
    def line(self, shape_):
        return LineFormat(shape_)

    @pytest.fixture
    def ln_(self, request):
        return instance_mock(request, CT_LineProperties)

    def ln_bldr(self, w):
        ln_bldr = an_ln().with_nsdecls()
        if w is not None:
            ln_bldr.with_w(w)
        return ln_bldr

    @pytest.fixture
    def shape_(self, request, ln_):
        shape_ = instance_mock(request, Shape)
        shape_.get_or_add_ln.return_value = ln_
        return shape_



# encoding: utf-8

"""Initializes lxml parser, particularly the custom element classes.

Also makes available a handful of functions that wrap its typical uses.
"""

from __future__ import absolute_import, division, print_function, unicode_literals

import os

from lxml import etree

from .ns import NamespacePrefixedTag


# configure etree XML parser -------------------------------
element_class_lookup = etree.ElementNamespaceClassLookup()
oxml_parser = etree.XMLParser(remove_blank_text=True, resolve_entities=False)
oxml_parser.set_element_class_lookup(element_class_lookup)


def parse_from_template(template_name):
    """
    Return an element loaded from the XML in the template file identified by
    *template_name*.
    """
    thisdir = os.path.split(__file__)[0]
    filename = os.path.join(thisdir, "..", "templates", "%s.xml" % template_name)
    with open(filename, "rb") as f:
        xml = f.read()
    return parse_xml(xml)


def parse_xml(xml):
    """
    Return root lxml element obtained by parsing XML character string in
    *xml*, which can be either a Python 2.x string or unicode.
    """
    root_element = etree.fromstring(xml, oxml_parser)
    return root_element


def register_element_cls(nsptagname, cls):
    """
    Register *cls* to be constructed when the oxml parser encounters an
    element having name *nsptag_name*. *nsptag_name* is a string of the form
    ``nspfx:tagroot``, e.g. ``'w:document'``.
    """
    nsptag = NamespacePrefixedTag(nsptagname)
    namespace = element_class_lookup.get_namespace(nsptag.nsuri)
    namespace[nsptag.local_part] = cls


from .action import CT_Hyperlink  # noqa: E402

register_element_cls("a:hlinkClick", CT_Hyperlink)
register_element_cls("a:hlinkHover", CT_Hyperlink)


from .chart.axis import (  # noqa: E402
    CT_AxisUnit,
    CT_CatAx,
    CT_ChartLines,
    CT_Crosses,
    CT_DateAx,
    CT_LblOffset,
    CT_Scaling,
    CT_TickLblPos,
    CT_TickMark,
    CT_ValAx,
)

register_element_cls("c:catAx", CT_CatAx)
register_element_cls("c:crosses", CT_Crosses)
register_element_cls("c:dateAx", CT_DateAx)
register_element_cls("c:lblOffset", CT_LblOffset)
register_element_cls("c:majorGridlines", CT_ChartLines)
register_element_cls("c:majorTickMark", CT_TickMark)
register_element_cls("c:majorUnit", CT_AxisUnit)
register_element_cls("c:minorTickMark", CT_TickMark)
register_element_cls("c:minorUnit", CT_AxisUnit)
register_element_cls("c:scaling", CT_Scaling)
register_element_cls("c:tickLblPos", CT_TickLblPos)
register_element_cls("c:valAx", CT_ValAx)


from .chart.chart import (  # noqa: E402
    CT_Chart,
    CT_ChartSpace,
    CT_ExternalData,
    CT_PlotArea,
    CT_Style,
)

register_element_cls("c:chart", CT_Chart)
register_element_cls("c:chartSpace", CT_ChartSpace)
register_element_cls("c:externalData", CT_ExternalData)
register_element_cls("c:plotArea", CT_PlotArea)
register_element_cls("c:style", CT_Style)


from .chart.datalabel import CT_DLbl, CT_DLblPos, CT_DLbls  # noqa: E402

register_element_cls("c:dLbl", CT_DLbl)
register_element_cls("c:dLblPos", CT_DLblPos)
register_element_cls("c:dLbls", CT_DLbls)


from .chart.legend import CT_Legend, CT_LegendPos  # noqa: E402

register_element_cls("c:legend", CT_Legend)
register_element_cls("c:legendPos", CT_LegendPos)


from .chart.marker import CT_Marker, CT_MarkerSize, CT_MarkerStyle  # noqa: E402

register_element_cls("c:marker", CT_Marker)
register_element_cls("c:size", CT_MarkerSize)
register_element_cls("c:symbol", CT_MarkerStyle)


from .chart.plot import (  # noqa: E402
    CT_Area3DChart,
    CT_AreaChart,
    CT_BarChart,
    CT_BarDir,
    CT_BubbleChart,
    CT_BubbleScale,
    CT_DoughnutChart,
    CT_GapAmount,
    CT_Grouping,
    CT_LineChart,
    CT_Overlap,
    CT_PieChart,
    CT_RadarChart,
    CT_ScatterChart,
)

register_element_cls("c:area3DChart", CT_Area3DChart)
register_element_cls("c:areaChart", CT_AreaChart)
register_element_cls("c:barChart", CT_BarChart)
register_element_cls("c:barDir", CT_BarDir)
register_element_cls("c:bubbleChart", CT_BubbleChart)
register_element_cls("c:bubbleScale", CT_BubbleScale)
register_element_cls("c:doughnutChart", CT_DoughnutChart)
register_element_cls("c:gapWidth", CT_GapAmount)
register_element_cls("c:grouping", CT_Grouping)
register_element_cls("c:lineChart", CT_LineChart)
register_element_cls("c:overlap", CT_Overlap)
register_element_cls("c:pieChart", CT_PieChart)
register_element_cls("c:radarChart", CT_RadarChart)
register_element_cls("c:scatterChart", CT_ScatterChart)


from .chart.series import (  # noqa: E402
    CT_AxDataSource,
    CT_DPt,
    CT_Lvl,
    CT_NumDataSource,
    CT_SeriesComposite,
    CT_StrVal_NumVal_Composite,
)

register_element_cls("c:bubbleSize", CT_NumDataSource)
register_element_cls("c:cat", CT_AxDataSource)
register_element_cls("c:dPt", CT_DPt)
register_element_cls("c:lvl", CT_Lvl)
register_element_cls("c:pt", CT_StrVal_NumVal_Composite)
register_element_cls("c:ser", CT_SeriesComposite)
register_element_cls("c:val", CT_NumDataSource)
register_element_cls("c:xVal", CT_NumDataSource)
register_element_cls("c:yVal", CT_NumDataSource)


from .chart.shared import (  # noqa: E402
    CT_Boolean,
    CT_Boolean_Explicit,
    CT_Double,
    CT_Layout,
    CT_LayoutMode,
    CT_ManualLayout,
    CT_NumFmt,
    CT_Title,
    CT_Tx,
    CT_UnsignedInt,
)

register_element_cls("c:autoTitleDeleted", CT_Boolean_Explicit)
register_element_cls("c:autoUpdate", CT_Boolean)
register_element_cls("c:bubble3D", CT_Boolean)
register_element_cls("c:crossAx", CT_UnsignedInt)
register_element_cls("c:crossesAt", CT_Double)
register_element_cls("c:date1904", CT_Boolean)
register_element_cls("c:delete", CT_Boolean)
register_element_cls("c:idx", CT_UnsignedInt)
register_element_cls("c:invertIfNegative", CT_Boolean_Explicit)
register_element_cls("c:layout", CT_Layout)
register_element_cls("c:manualLayout", CT_ManualLayout)
register_element_cls("c:max", CT_Double)
register_element_cls("c:min", CT_Double)
register_element_cls("c:numFmt", CT_NumFmt)
register_element_cls("c:order", CT_UnsignedInt)
register_element_cls("c:overlay", CT_Boolean_Explicit)
register_element_cls("c:ptCount", CT_UnsignedInt)
register_element_cls("c:showCatName", CT_Boolean_Explicit)
register_element_cls("c:showLegendKey", CT_Boolean_Explicit)
register_element_cls("c:showPercent", CT_Boolean_Explicit)
register_element_cls("c:showSerName", CT_Boolean_Explicit)
register_element_cls("c:showVal", CT_Boolean_Explicit)
register_element_cls("c:smooth", CT_Boolean)
register_element_cls("c:title", CT_Title)
register_element_cls("c:tx", CT_Tx)
register_element_cls("c:varyColors", CT_Boolean)
register_element_cls("c:x", CT_Double)
register_element_cls("c:xMode", CT_LayoutMode)


from .coreprops import CT_CoreProperties  # noqa: E402

register_element_cls("cp:coreProperties", CT_CoreProperties)


from .dml.color import (  # noqa: E402
    CT_Color,
    CT_HslColor,
    CT_Percentage,
    CT_PresetColor,
    CT_SchemeColor,
    CT_ScRgbColor,
    CT_SRgbColor,
    CT_SystemColor,
)

register_element_cls("a:bgClr", CT_Color)
register_element_cls("a:fgClr", CT_Color)
register_element_cls("a:hslClr", CT_HslColor)
register_element_cls("a:lumMod", CT_Percentage)
register_element_cls("a:lumOff", CT_Percentage)
register_element_cls("a:prstClr", CT_PresetColor)
register_element_cls("a:schemeClr", CT_SchemeColor)
register_element_cls("a:scrgbClr", CT_ScRgbColor)
register_element_cls("a:srgbClr", CT_SRgbColor)
register_element_cls("a:sysClr", CT_SystemColor)


from .dml.fill import (  # noqa: E402
    CT_Blip,
    CT_BlipFillProperties,
    CT_GradientFillProperties,
    CT_GradientStop,
    CT_GradientStopList,
    CT_GroupFillProperties,
    CT_LinearShadeProperties,
    CT_NoFillProperties,
    CT_PatternFillProperties,
    CT_RelativeRect,
    CT_SolidColorFillProperties,
)

register_element_cls("a:blip", CT_Blip)
register_element_cls("a:blipFill", CT_BlipFillProperties)
register_element_cls("a:gradFill", CT_GradientFillProperties)
register_element_cls("a:grpFill", CT_GroupFillProperties)
register_element_cls("a:gs", CT_GradientStop)
register_element_cls("a:gsLst", CT_GradientStopList)
register_element_cls("a:lin", CT_LinearShadeProperties)
register_element_cls("a:noFill", CT_NoFillProperties)
register_element_cls("a:pattFill", CT_PatternFillProperties)
register_element_cls("a:solidFill", CT_SolidColorFillProperties)
register_element_cls("a:srcRect", CT_RelativeRect)


from .dml.line import CT_PresetLineDashProperties  # noqa: E402

register_element_cls("a:prstDash", CT_PresetLineDashProperties)


from .presentation import (  # noqa: E402
    CT_Presentation,
    CT_SlideId,
    CT_SlideIdList,
    CT_SlideMasterIdList,
    CT_SlideMasterIdListEntry,
    CT_SlideSize,
)

register_element_cls("p:presentation", CT_Presentation)
register_element_cls("p:sldId", CT_SlideId)
register_element_cls("p:sldIdLst", CT_SlideIdList)
register_element_cls("p:sldMasterId", CT_SlideMasterIdListEntry)
register_element_cls("p:sldMasterIdLst", CT_SlideMasterIdList)
register_element_cls("p:sldSz", CT_SlideSize)


from .shapes.autoshape import (  # noqa: E402
    CT_AdjPoint2D,
    CT_CustomGeometry2D,
    CT_GeomGuide,
    CT_GeomGuideList,
    CT_NonVisualDrawingShapeProps,
    CT_Path2D,
    CT_Path2DClose,
    CT_Path2DLineTo,
    CT_Path2DList,
    CT_Path2DMoveTo,
    CT_PresetGeometry2D,
    CT_Shape,
    CT_ShapeNonVisual,
)

register_element_cls("a:avLst", CT_GeomGuideList)
register_element_cls("a:custGeom", CT_CustomGeometry2D)
register_element_cls("a:gd", CT_GeomGuide)
register_element_cls("a:close", CT_Path2DClose)
register_element_cls("a:lnTo", CT_Path2DLineTo)
register_element_cls("a:moveTo", CT_Path2DMoveTo)
register_element_cls("a:path", CT_Path2D)
register_element_cls("a:pathLst", CT_Path2DList)
register_element_cls("a:prstGeom", CT_PresetGeometry2D)
register_element_cls("a:pt", CT_AdjPoint2D)
register_element_cls("p:cNvSpPr", CT_NonVisualDrawingShapeProps)
register_element_cls("p:nvSpPr", CT_ShapeNonVisual)
register_element_cls("p:sp", CT_Shape)


from .shapes.connector import (  # noqa: E402
    CT_Connection,
    CT_Connector,
    CT_ConnectorNonVisual,
    CT_NonVisualConnectorProperties,
)

register_element_cls("a:endCxn", CT_Connection)
register_element_cls("a:stCxn", CT_Connection)
register_element_cls("p:cNvCxnSpPr", CT_NonVisualConnectorProperties)
register_element_cls("p:cxnSp", CT_Connector)
register_element_cls("p:nvCxnSpPr", CT_ConnectorNonVisual)


from .shapes.graphfrm import (  # noqa: E402
    CT_GraphicalObject,
    CT_GraphicalObjectData,
    CT_GraphicalObjectFrame,
    CT_GraphicalObjectFrameNonVisual,
    CT_OleObject,
)

register_element_cls("a:graphic", CT_GraphicalObject)
register_element_cls("a:graphicData", CT_GraphicalObjectData)
register_element_cls("p:graphicFrame", CT_GraphicalObjectFrame)
register_element_cls("p:nvGraphicFramePr", CT_GraphicalObjectFrameNonVisual)
register_element_cls("p:oleObj", CT_OleObject)


from .shapes.groupshape import (  # noqa: E402
    CT_GroupShape,
    CT_GroupShapeNonVisual,
    CT_GroupShapeProperties,
)

register_element_cls("p:grpSp", CT_GroupShape)
register_element_cls("p:grpSpPr", CT_GroupShapeProperties)
register_element_cls("p:nvGrpSpPr", CT_GroupShapeNonVisual)
register_element_cls("p:spTree", CT_GroupShape)


from .shapes.picture import CT_Picture, CT_PictureNonVisual  # noqa: E402

register_element_cls("p:blipFill", CT_BlipFillProperties)
register_element_cls("p:nvPicPr", CT_PictureNonVisual)
register_element_cls("p:pic", CT_Picture)


from .shapes.shared import (  # noqa: E402
    CT_ApplicationNonVisualDrawingProps,
    CT_LineProperties,
    CT_NonVisualDrawingProps,
    CT_Placeholder,
    CT_Point2D,
    CT_PositiveSize2D,
    CT_ShapeProperties,
    CT_Transform2D,
)

register_element_cls("a:chExt", CT_PositiveSize2D)
register_element_cls("a:chOff", CT_Point2D)
register_element_cls("a:ext", CT_PositiveSize2D)
register_element_cls("a:ln", CT_LineProperties)
register_element_cls("a:off", CT_Point2D)
register_element_cls("a:xfrm", CT_Transform2D)
register_element_cls("c:spPr", CT_ShapeProperties)
register_element_cls("p:cNvPr", CT_NonVisualDrawingProps)
register_element_cls("p:nvPr", CT_ApplicationNonVisualDrawingProps)
register_element_cls("p:ph", CT_Placeholder)
register_element_cls("p:spPr", CT_ShapeProperties)
register_element_cls("p:xfrm", CT_Transform2D)


from .slide import (  # noqa: E402
    CT_Background,
    CT_BackgroundProperties,
    CT_CommonSlideData,
    CT_NotesMaster,
    CT_NotesSlide,
    CT_Slide,
    CT_SlideLayout,
    CT_SlideLayoutIdList,
    CT_SlideLayoutIdListEntry,
    CT_SlideMaster,
    CT_SlideTiming,
    CT_TimeNodeList,
    CT_TLMediaNodeVideo,
)

register_element_cls("p:bg", CT_Background)
register_element_cls("p:bgPr", CT_BackgroundProperties)
register_element_cls("p:childTnLst", CT_TimeNodeList)
register_element_cls("p:cSld", CT_CommonSlideData)
register_element_cls("p:notes", CT_NotesSlide)
register_element_cls("p:notesMaster", CT_NotesMaster)
register_element_cls("p:sld", CT_Slide)
register_element_cls("p:sldLayout", CT_SlideLayout)
register_element_cls("p:sldLayoutId", CT_SlideLayoutIdListEntry)
register_element_cls("p:sldLayoutIdLst", CT_SlideLayoutIdList)
register_element_cls("p:sldMaster", CT_SlideMaster)
register_element_cls("p:timing", CT_SlideTiming)
register_element_cls("p:video", CT_TLMediaNodeVideo)


from .table import (  # noqa: E402
    CT_Table,
    CT_TableCell,
    CT_TableCellProperties,
    CT_TableCol,
    CT_TableGrid,
    CT_TableProperties,
    CT_TableRow,
)

register_element_cls("a:gridCol", CT_TableCol)
register_element_cls("a:tbl", CT_Table)
register_element_cls("a:tblGrid", CT_TableGrid)
register_element_cls("a:tblPr", CT_TableProperties)
register_element_cls("a:tc", CT_TableCell)
register_element_cls("a:tcPr", CT_TableCellProperties)
register_element_cls("a:tr", CT_TableRow)


from .text import (  # noqa: E402
    CT_RegularTextRun,
    CT_TextBody,
    CT_TextBodyProperties,
    CT_TextCharacterProperties,
    CT_TextField,
    CT_TextFont,
    CT_TextLineBreak,
    CT_TextNormalAutofit,
    CT_TextParagraph,
    CT_TextParagraphProperties,
    CT_TextSpacing,
    CT_TextSpacingPercent,
    CT_TextSpacingPoint,
)

register_element_cls("a:bodyPr", CT_TextBodyProperties)
register_element_cls("a:br", CT_TextLineBreak)
register_element_cls("a:defRPr", CT_TextCharacterProperties)
register_element_cls("a:endParaRPr", CT_TextCharacterProperties)
register_element_cls("a:fld", CT_TextField)
register_element_cls("a:latin", CT_TextFont)
register_element_cls("a:lnSpc", CT_TextSpacing)
register_element_cls("a:normAutofit", CT_TextNormalAutofit)
register_element_cls("a:r", CT_RegularTextRun)
register_element_cls("a:p", CT_TextParagraph)
register_element_cls("a:pPr", CT_TextParagraphProperties)
register_element_cls("c:rich", CT_TextBody)
register_element_cls("a:rPr", CT_TextCharacterProperties)
register_element_cls("a:spcAft", CT_TextSpacing)
register_element_cls("a:spcBef", CT_TextSpacing)
register_element_cls("a:spcPct", CT_TextSpacingPercent)
register_element_cls("a:spcPts", CT_TextSpacingPoint)
register_element_cls("a:txBody", CT_TextBody)
register_element_cls("c:txPr", CT_TextBody)
register_element_cls("p:txBody", CT_TextBody)


from .theme import CT_OfficeStyleSheet  # noqa: E402

register_element_cls("a:theme", CT_OfficeStyleSheet)



# encoding: utf-8

"""ImagePart and related objects."""

from __future__ import absolute_import, division, print_function, unicode_literals

import hashlib
import os

try:
    from PIL import Image as PIL_Image
except ImportError:
    import Image as PIL_Image

from ..compat import BytesIO, is_string
from ..opc.package import Part
from ..opc.spec import image_content_types
from ..util import lazyproperty


class ImagePart(Part):
    """
    An image part, generally having a partname matching the regex
    ``ppt/media/image[1-9][0-9]*.*``.
    """

    def __init__(self, partname, content_type, blob, package, filename=None):
        super(ImagePart, self).__init__(partname, content_type, blob, package)
        self._filename = filename

    @classmethod
    def load(cls, partname, content_type, blob, package):
        return cls(partname, content_type, blob, package)

    @classmethod
    def new(cls, package, image):
        """
        Return a new |ImagePart| instance containing *image*, which is an
        |Image| object.
        """
        partname = package.next_image_partname(image.ext)
        return cls(partname, image.content_type, image.blob, package, image.filename)

    @property
    def desc(self):
        """
        The filename associated with this image, either the filename of
        the original image or a generic name of the form ``image.ext``
        where ``ext`` is appropriate to the image file format, e.g.
        ``'jpg'``. An image created using a path will have that filename; one
        created with a file-like object will have a generic name.
        """
        # return generic filename if original filename is unknown
        if self._filename is None:
            return "image.%s" % self.ext
        return self._filename

    @property
    def ext(self):
        """
        Return file extension for this image e.g. ``'png'``.
        """
        return self.partname.ext

    @property
    def image(self):
        """
        An |Image| object containing the image in this image part.
        """
        return Image(self.blob, self.desc)

    def scale(self, scaled_cx, scaled_cy):
        """
        Return scaled image dimensions in EMU based on the combination of
        parameters supplied. If *scaled_cx* and *scaled_cy* are both |None|,
        the native image size is returned. If neither *scaled_cx* nor
        *scaled_cy* is |None|, their values are returned unchanged. If
        a value is provided for either *scaled_cx* or *scaled_cy* and the
        other is |None|, the missing value is calculated such that the
        image's aspect ratio is preserved.
        """
        image_cx, image_cy = self._native_size

        if scaled_cx is None and scaled_cy is None:
            scaled_cx = image_cx
            scaled_cy = image_cy
        elif scaled_cx is None:
            scaling_factor = float(scaled_cy) / float(image_cy)
            scaled_cx = int(round(image_cx * scaling_factor))
        elif scaled_cy is None:
            scaling_factor = float(scaled_cx) / float(image_cx)
            scaled_cy = int(round(image_cy * scaling_factor))

        return scaled_cx, scaled_cy

    @lazyproperty
    def sha1(self):
        """
        The SHA1 hash digest for the image binary of this image part, like:
        ``'1be010ea47803b00e140b852765cdf84f491da47'``.
        """
        return hashlib.sha1(self._blob).hexdigest()

    @property
    def _dpi(self):
        """
        A (horz_dpi, vert_dpi) 2-tuple (ints) representing the dots-per-inch
        property of this image.
        """
        image = Image.from_blob(self.blob)
        return image.dpi

    @property
    def _native_size(self):
        """
        A (width, height) 2-tuple representing the native dimensions of the
        image in EMU, calculated based on the image DPI value, if present,
        assuming 72 dpi as a default.
        """
        EMU_PER_INCH = 914400
        horz_dpi, vert_dpi = self._dpi
        width_px, height_px = self._px_size

        width = EMU_PER_INCH * width_px / horz_dpi
        height = EMU_PER_INCH * height_px / vert_dpi

        return width, height

    @property
    def _px_size(self):
        """
        A (width, height) 2-tuple representing the dimensions of this image
        in pixels.
        """
        image = Image.from_blob(self.blob)
        return image.size


class Image(object):
    """
    Immutable value object representing an image such as a JPEG, PNG, or GIF.
    """

    def __init__(self, blob, filename):
        super(Image, self).__init__()
        self._blob = blob
        self._filename = filename

    @classmethod
    def from_blob(cls, blob, filename=None):
        """
        Return a new |Image| object loaded from the image binary in *blob*.
        """
        return cls(blob, filename)

    @classmethod
    def from_file(cls, image_file):
        """
        Return a new |Image| object loaded from *image_file*, which can be
        either a path (string) or a file-like object.
        """
        if is_string(image_file):
            # treat image_file as a path
            with open(image_file, "rb") as f:
                blob = f.read()
            filename = os.path.basename(image_file)
        else:
            # assume image_file is a file-like object
            # ---reposition file cursor if it has one---
            if callable(getattr(image_file, "seek")):
                image_file.seek(0)
            blob = image_file.read()
            filename = None

        return cls.from_blob(blob, filename)

    @property
    def blob(self):
        """
        The binary image bytestream of this image.
        """
        return self._blob

    @lazyproperty
    def content_type(self):
        """
        MIME-type of this image, e.g. ``'image/jpeg'``.
        """
        return image_content_types[self.ext]

    @lazyproperty
    def dpi(self):
        """
        A (horz_dpi, vert_dpi) 2-tuple specifying the dots-per-inch
        resolution of this image. A default value of (72, 72) is used if the
        dpi is not specified in the image file.
        """

        def int_dpi(dpi):
            """
            Return an integer dots-per-inch value corresponding to *dpi*. If
            *dpi* is |None|, a non-numeric type, less than 1 or greater than
            2048, 72 is returned.
            """
            try:
                int_dpi = int(round(float(dpi)))
                if int_dpi < 1 or int_dpi > 2048:
                    int_dpi = 72
            except (TypeError, ValueError):
                int_dpi = 72
            return int_dpi

        def normalize_pil_dpi(pil_dpi):
            """
            Return a (horz_dpi, vert_dpi) 2-tuple corresponding to *pil_dpi*,
            the value for the 'dpi' key in the ``info`` dict of a PIL image.
            If the 'dpi' key is not present or contains an invalid value,
            ``(72, 72)`` is returned.
            """
            if isinstance(pil_dpi, tuple):
                return (int_dpi(pil_dpi[0]), int_dpi(pil_dpi[1]))
            return (72, 72)

        return normalize_pil_dpi(self._pil_props[2])

    @lazyproperty
    def ext(self):
        """
        Canonical file extension for this image e.g. ``'png'``. The returned
        extension is all lowercase and is the canonical extension for the
        content type of this image, regardless of what extension may have
        been used in its filename, if any.
        """
        ext_map = {
            "BMP": "bmp",
            "GIF": "gif",
            "JPEG": "jpg",
            "PNG": "png",
            "TIFF": "tiff",
            "WMF": "wmf",
        }
        format = self._format
        if format not in ext_map:
            tmpl = "unsupported image format, expected one of: %s, got '%s'"
            raise ValueError(tmpl % (ext_map.keys(), format))
        return ext_map[format]

    @property
    def filename(self):
        """
        The filename from the path from which this image was loaded, if
        loaded from the filesystem. |None| if no filename was used in
        loading, such as when loaded from an in-memory stream.
        """
        return self._filename

    @lazyproperty
    def sha1(self):
        """
        SHA1 hash digest of the image blob
        """
        return hashlib.sha1(self._blob).hexdigest()

    @lazyproperty
    def size(self):
        """
        A (width, height) 2-tuple specifying the dimensions of this image in
        pixels.
        """
        return self._pil_props[1]

    @property
    def _format(self):
        """
        The PIL Image format of this image, e.g. 'PNG'.
        """
        return self._pil_props[0]

    @lazyproperty
    def _pil_props(self):
        """
        A tuple containing useful image properties extracted from this image
        using Pillow (Python Imaging Library, or 'PIL').
        """
        stream = BytesIO(self._blob)
        pil_image = PIL_Image.open(stream)
        format = pil_image.format
        width_px, height_px = pil_image.size
        dpi = pil_image.info.get("dpi")
        stream.close()
        return (format, (width_px, height_px), dpi)



#!/usr/bin/env python

import os
import re

from setuptools import find_packages, setup


def ascii_bytes_from(path, *paths):
    """
    Return the ASCII characters in the file specified by *path* and *paths*.
    The file path is determined by concatenating *path* and any members of
    *paths* with a directory separator in between.
    """
    file_path = os.path.join(path, *paths)
    with open(file_path) as f:
        ascii_bytes = f.read()
    return ascii_bytes


# read required text from files
thisdir = os.path.dirname(__file__)
init_py = ascii_bytes_from(thisdir, "pptx", "__init__.py")
readme = ascii_bytes_from(thisdir, "README.rst")
history = ascii_bytes_from(thisdir, "HISTORY.rst")
license = ascii_bytes_from(thisdir, "LICENSE")

# Read the version from pptx.__version__ without importing the package
# (and thus attempting to import packages it depends on that may not be
# installed yet)
version = re.search(r'__version__ = "([^"]+)"', init_py).group(1)


NAME = "iroin-python-pptx"
VERSION = version
DESCRIPTION = "Generate and manipulate Open XML PowerPoint (.pptx) files"
KEYWORDS = "powerpoint ppt pptx office open xml"
AUTHOR = "iroin"
AUTHOR_EMAIL = "admin@iroin.io"
URL = "https://github.com/iroin/python-pptx"
LICENSE = license
PACKAGES = find_packages(exclude=["tests", "tests.*"])
PACKAGE_DATA = {"pptx": ["templates/*"]}

INSTALL_REQUIRES = ["lxml>=3.1.0", "Pillow>=3.3.2", "XlsxWriter>=0.5.7"]

TEST_SUITE = "tests"
TESTS_REQUIRE = ["behave", "mock", "pyparsing>=2.0.1", "pytest"]

CLASSIFIERS = [
    "Development Status :: 4 - Beta",
    "Environment :: Console",
    "Intended Audience :: Developers",
    "License :: OSI Approved :: MIT License",
    "Operating System :: OS Independent",
    "Programming Language :: Python",
    "Programming Language :: Python :: 2",
    "Programming Language :: Python :: 2.7",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.6",
    "Topic :: Office/Business :: Office Suites",
    "Topic :: Software Development :: Libraries",
]

LONG_DESCRIPTION = "Some Text"

ZIP_SAFE = False

params = {
    "name": NAME,
    "version": VERSION,
    "description": DESCRIPTION,
    "keywords": KEYWORDS,
    "long_description": LONG_DESCRIPTION,
    "author": AUTHOR,
    "author_email": AUTHOR_EMAIL,
    "url": URL,
    "license": LICENSE,
    "packages": PACKAGES,
    "package_data": PACKAGE_DATA,
    "install_requires": INSTALL_REQUIRES,
    "tests_require": TESTS_REQUIRE,
    "test_suite": TEST_SUITE,
    "classifiers": CLASSIFIERS,
    "zip_safe": ZIP_SAFE,
}

setup(**params)



# encoding: utf-8

"""
Test suite for pptx.oxml.simpletypes module, which contains simple type class
definitions. A simple type in this context corresponds to an
``<xsd:simpleType>`` definition in the XML schema and provides data
validation and type conversion services for use by xmlchemy.
"""

from __future__ import absolute_import, print_function

import pytest

from pptx.oxml.simpletypes import (
    BaseIntType,
    BaseSimpleType,
    ST_Coordinate,
    ST_HexColorRGB,
    ST_Percentage,
)

from ..unitutil.mock import method_mock, instance_mock


class DescribeBaseSimpleType(object):
    def it_can_convert_attr_value_to_python_type(self, from_xml_fixture):
        SimpleType, str_value_, py_value_ = from_xml_fixture
        py_value = SimpleType.from_xml(str_value_)
        SimpleType.convert_from_xml.assert_called_once_with(str_value_)
        assert py_value is py_value_

    def it_can_convert_python_value_to_string(self, to_xml_fixture):
        SimpleType, py_value_, str_value_ = to_xml_fixture
        str_value = SimpleType.to_xml(py_value_)
        SimpleType.validate.assert_called_once_with(py_value_)
        SimpleType.convert_to_xml.assert_called_once_with(py_value_)
        assert str_value is str_value_

    def it_can_validate_a_value_as_a_python_int(self, valid_int_fixture):
        value, expected_exception = valid_int_fixture
        if expected_exception is None:
            BaseSimpleType.validate_int(value)
        else:
            with pytest.raises(expected_exception):
                BaseSimpleType.validate_int(value)

    def it_can_validate_a_value_as_a_python_string(self, valid_str_fixture):
        value, expected_exception = valid_str_fixture
        if expected_exception is None:
            BaseSimpleType.validate_string(value)
        else:
            with pytest.raises(expected_exception):
                BaseSimpleType.validate_string(value)

    # fixtures -------------------------------------------------------

    @pytest.fixture
    def from_xml_fixture(self, request, str_value_, py_value_, convert_from_xml_):
        return ST_SimpleType, str_value_, py_value_

    @pytest.fixture
    def to_xml_fixture(
        self, request, py_value_, str_value_, convert_to_xml_, validate_
    ):
        return ST_SimpleType, py_value_, str_value_

    @pytest.fixture(
        params=[
            (42, None),
            (0, None),
            (-42, None),
            ("42", TypeError),
            (None, TypeError),
            (42.42, TypeError),
        ]
    )
    def valid_int_fixture(self, request):
        value, expected_exception = request.param
        return value, expected_exception

    @pytest.fixture(
        params=[
            ("foobar", None),
            ("", None),
            (" foo ", None),
            (("foo",), TypeError),
            (42, TypeError),
            (None, TypeError),
            (42.42, TypeError),
        ]
    )
    def valid_str_fixture(self, request):
        value, expected_exception = request.param
        return value, expected_exception

    # fixture components ---------------------------------------------

    @pytest.fixture
    def convert_from_xml_(self, request, py_value_):
        return method_mock(
            request, ST_SimpleType, "convert_from_xml", return_value=py_value_
        )

    @pytest.fixture
    def convert_to_xml_(self, request, str_value_):
        return method_mock(
            request, ST_SimpleType, "convert_to_xml", return_value=str_value_
        )

    @pytest.fixture
    def py_value_(self, request):
        return instance_mock(request, int)

    @pytest.fixture
    def str_value_(self, request):
        return instance_mock(request, str)

    @pytest.fixture
    def validate_(self, request):
        return method_mock(request, ST_SimpleType, "validate")


class DescribeBaseIntType(object):
    def it_can_convert_a_string_to_an_int(self, from_xml_fixture):
        str_value, expected_value, expected_exception = from_xml_fixture
        if expected_exception is None:
            value = BaseIntType.convert_from_xml(str_value)
            assert value == expected_value
        else:
            with pytest.raises(expected_exception):
                BaseIntType.convert_from_xml(str_value)

    def it_can_convert_an_int_to_a_string(self, to_xml_fixture):
        value, expected_str_value = to_xml_fixture
        str_value = BaseIntType.convert_to_xml(value)
        assert str_value == expected_str_value

    # fixtures -------------------------------------------------------

    @pytest.fixture(
        params=[
            ("42", 42, None),
            ("-42", -42, None),
            ("-0042", -42, None),
            ("", None, ValueError),
            ("foo", None, ValueError),
            ("42.42", None, ValueError),
            ("0x0a3", None, ValueError),
            (None, None, TypeError),
        ]
    )
    def from_xml_fixture(self, request):
        str_value, expected_value, expected_exception = request.param
        return str_value, expected_value, expected_exception

    @pytest.fixture(params=[(42, "42"), (-42, "-42"), (0x2A, "42")])
    def to_xml_fixture(self, request):
        value, expected_str_value = request.param
        return value, expected_str_value


class DescribeST_Coordinate(object):
    def it_can_convert_from_ST_UniversalMeasure(self, univ_meas_fixture):
        """
        ST_UniversalMeasure allows strings line '-034.56pt'.
        """
        str_value, expected_value = univ_meas_fixture
        assert ST_Coordinate.convert_from_xml(str_value) == expected_value

    # fixtures -------------------------------------------------------

    @pytest.fixture(
        params=[
            ("1.2in", 1097280),
            ("42mm", 1512000),
            ("0024cm", 8640000),
            ("-42pt", -533400),
            ("-036.214pc", -5519014),
            ("0pi", 0),
        ]
    )
    def univ_meas_fixture(self, request):
        str_value, expected_value = request.param
        return str_value, expected_value


class DescribeST_HexColorRGB(object):
    def it_can_validate_a_hex_RGB_string(self, valid_fixture):
        str_value, exception = valid_fixture
        if exception is None:
            try:
                ST_HexColorRGB.validate(str_value)
            except ValueError:
                raise AssertionError("string '%s' did not validate" % str_value)
        else:
            with pytest.raises(exception):
                ST_HexColorRGB.validate(str_value)

    def it_converts_RGB_string_to_upper_case(self, to_xml_fixture):
        value, expected_value = to_xml_fixture
        _value = ST_HexColorRGB.convert_to_xml(value)
        assert _value == expected_value

    # fixtures -------------------------------------------------------

    @pytest.fixture(
        params=[("deadbf", "DEADBF"), ("012345", "012345"), ("0a1b3c", "0A1B3C")]
    )
    def to_xml_fixture(self, request):
        value, expected_value = request.param
        return value, expected_value

    @pytest.fixture(
        params=[
            ("012345", None),
            ("ABCDEF", None),
            ("deadbf", None),
            ("0A1B3C", None),
            (None, TypeError),  # must be string
            (123456, TypeError),  # must be string
            ("F00BAR", ValueError),  # includes non-hex character 'R'
            ("F00b", ValueError),  # too short, must be 6 characters
        ]
    )
    def valid_fixture(self, request):
        str_value, exception = request.param
        return str_value, exception


class DescribeST_Percentage(object):
    def it_can_convert_from_xml_percent_literals(self, percent_fixture):
        str_value, expected_value = percent_fixture
        value = ST_Percentage.convert_from_xml(str_value)
        assert abs(value - expected_value) < 0.000001

    # fixtures -------------------------------------------------------

    @pytest.fixture(
        params=[
            ("12.34%", 0.1234),
            ("42%", 0.42),
            ("024%", 0.24),
            ("-42%", -0.42),
            ("-036.214%", -0.36214),
            ("0%", 0.0),
        ]
    )
    def percent_fixture(self, request):
        str_value, expected_value = request.param
        return str_value, expected_value


# --------------------------------------------------------------------
# static fixture
# --------------------------------------------------------------------


class ST_SimpleType(BaseSimpleType):
    @classmethod
    def convert_from_xml(cls, str_value):
        return 666

    @classmethod
    def convert_to_xml(cls, value):
        return "666"

    @classmethod
    def validate(cls, value):
        pass
